"""Tests for PyTorchDataset classes."""
from typing import Union

from PIL import Image
import numpy as np
import pandas as pd
import pytest
import torch

from bitfount.backends.pytorch.data.datasets import _PyTorchDataset
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.datastructure import DataStructure
from bitfount.data.schema import BitfountSchema
from bitfount.data.types import DataSplit, SemanticType
from tests.utils.helper import TABLE_NAME, backend_test, create_dataset, unit_test


@backend_test
@unit_test
class TestPyTorchDataset:
    """Tests for PyTorchTabularDataset class."""

    def test_len_tab_data(self, tabular_dataset: _PyTorchDataset) -> None:
        """Tests tabular dataset __len__ method."""
        assert tabular_dataset.datasource._train_idxs is not None
        assert len(tabular_dataset) == len(tabular_dataset.datasource._train_idxs)

    def test_len_image_data(self, image_dataset: _PyTorchDataset) -> None:
        """Tests image dataset __len__ method."""
        assert image_dataset.datasource._train_idxs is not None
        assert len(image_dataset) == len(image_dataset.datasource._train_idxs)

    def test_len_image_tab_data(self, image_tab_dataset: _PyTorchDataset) -> None:
        """Tests mixed dataset __len__ method."""
        assert image_tab_dataset.datasource._train_idxs is not None
        assert len(image_tab_dataset) == len(image_tab_dataset.datasource._train_idxs)

    def test_len_multiimage_data(self, multiimage_dataset: _PyTorchDataset) -> None:
        """Tests multi-image dataset __len__ method."""
        assert multiimage_dataset.datasource._train_idxs is not None
        assert len(multiimage_dataset) == len(multiimage_dataset.datasource._train_idxs)

    @pytest.mark.parametrize("idx", [0, 42, 2048, torch.tensor(3199)])
    def test_idx_tab_data(
        self, idx: Union[int, torch.Tensor], tabular_dataset: _PyTorchDataset
    ) -> None:
        """Tests indexing (incl. tensors) returns the expected formats of data."""
        assert isinstance(tabular_dataset[idx], tuple)
        assert len(tabular_dataset[idx]) == 2  # split into x,y
        assert len(tabular_dataset[idx][0]) == 2  # split into tabular, support
        assert len(tabular_dataset[idx][0][0]) == 13  # training cols  check
        assert len(tabular_dataset[idx][0][1]) == 2  # support cols check
        assert len([tabular_dataset[idx][1]]) == 1  # y check

    @pytest.mark.parametrize("idx", [0, 42, 2048, torch.tensor(3199)])
    def test_idx_img_data(self, idx: int, image_dataset: _PyTorchDataset) -> None:
        """Tests indexing returns the expected formats of data."""
        assert isinstance(image_dataset[idx], tuple)
        assert len(image_dataset[idx]) == 2  # split into x,y
        assert len(image_dataset[idx][0]) == 2  # split into image, support
        assert len(image_dataset[idx][0][1]) == 2  # support cols check
        assert len([image_dataset[idx][1]]) == 1  # y check

    @pytest.mark.parametrize("idx", [0, 42, 2048, torch.tensor(3199)])
    def test_idx_img_tab_data(
        self, idx: int, image_tab_dataset: _PyTorchDataset
    ) -> None:
        """Tests indexing returns the expected formats of data."""
        assert isinstance(image_tab_dataset[idx], tuple)
        assert len(image_tab_dataset[idx]) == 2  # split into x,y
        assert len(image_tab_dataset[idx][0]) == 3  # split into tab, image, support
        assert len(image_tab_dataset[idx][0][0]) == 13  # tabular cols  check
        # support cols check; mypy doesn't know that there are greater than 2 elements
        assert len(image_tab_dataset[idx][0][2]) == 2  # type: ignore[misc] # Reason: see above # noqa: B950
        assert len([image_tab_dataset[idx][1]]) == 1  # y check

    @pytest.mark.parametrize("idx", [0, 42, 2048, torch.tensor(3199)])
    def test_idx_multiimg_data(
        self, idx: int, multiimage_dataset: _PyTorchDataset
    ) -> None:
        """Tests indexing returns the expected formats of data."""
        assert isinstance(multiimage_dataset[idx], tuple)
        assert len(multiimage_dataset[idx]) == 2  # split into x,y
        assert len(multiimage_dataset[idx][0]) == 2  # split into image, support
        assert isinstance(multiimage_dataset[idx][0][0], tuple)
        assert len(multiimage_dataset[idx][0][0]) == 2  # image cols check
        assert len(multiimage_dataset[idx][0][1]) == 2  # support cols check
        assert len([multiimage_dataset[idx][1]]) == 1  # y check

    @pytest.mark.parametrize("idx", [0, 42, 2048, torch.tensor(3199)])
    def test_idx_img_tab_category(
        self, idx: int, image_tab_dataset: _PyTorchDataset
    ) -> None:
        """Tests indexing with categories gives expected data formats."""
        target = "TARGET"
        data = create_dataset(image=True, multihead=True)
        datasource = DataFrameSource(data, image_col=["image"])
        schema = BitfountSchema()
        schema.add_datasource_tables(
            datasource,
            force_stypes={
                TABLE_NAME: {"categorical": ["category"], "image": ["image"]}
            },
            table_name=TABLE_NAME,
        )
        datasource.load_data()
        datasource.data = datasource.data.drop(
            columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
        )
        datastructure = DataStructure(
            target=target, multihead_col="category", multihead_size=2, table=TABLE_NAME
        )
        datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
        dataset = _PyTorchDataset(
            datasource=datasource,
            target=target,
            multihead_col="category",
            selected_cols_semantic_types=datastructure.selected_cols_w_types,
            selected_cols=datastructure.selected_cols,
            schema=schema.get_table_schema(TABLE_NAME),
            data_split=DataSplit.TRAIN,
        )
        assert isinstance(dataset[idx], tuple)
        assert len(dataset[idx]) == 2  # split into x,y
        assert len(dataset[idx][0]) == 3  # split into tab, image, support
        assert (
            len(dataset[idx][0][0]) == 14
        )  # tabular cols check (multihead_col included)
        # support cols check; mypy doesn't know that there are greater than 2 elements
        assert len(dataset[idx][0][2]) == 3  # type: ignore[misc] # Reason: see above # noqa: B950
        assert len([dataset[idx][1]]) == 1  # y check

    def test_dataset_works_only_with_continuous_features(
        self, dataframe: pd.DataFrame
    ) -> None:
        """Test no errors are raised if the dataset only has continuous features."""
        datasource = DataFrameSource(dataframe.loc[:, ["A", "B", "TARGET"]])
        datasource.load_data()
        schema = BitfountSchema()
        schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
        datastructure = DataStructure(
            target=["TARGET"],
            table=TABLE_NAME,
        )
        datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
        dataset = _PyTorchDataset(
            datasource=datasource,
            target=["TARGET"],
            selected_cols_semantic_types=datastructure.selected_cols_w_types,
            selected_cols=datastructure.selected_cols,
            schema=schema.get_table_schema(TABLE_NAME),
            data_split=DataSplit.TRAIN,
        )
        assert "categorical" not in schema.tables[0].features
        idx = 10
        assert isinstance(dataset[idx], tuple)
        assert len(dataset[idx]) == 2  # split into x,y
        assert len(dataset[idx][0]) == 2  # split into tabular, support
        assert len(dataset[idx][0][0]) == 2  # training cols  check
        assert len(dataset[idx][0][1]) == 2  # support cols check
        assert len([dataset[idx][1]]) == 1  # y check

    def test_dataset_works_without_target(self, dataframe: pd.DataFrame) -> None:
        """Test no errors are raised if the dataset has no target."""
        datasource = DataFrameSource(dataframe.loc[:, ["A", "B"]])
        datasource.load_data()
        schema = BitfountSchema()
        schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
        datastructure = DataStructure(table=TABLE_NAME)
        datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
        dataset = _PyTorchDataset(
            datasource=datasource,
            selected_cols_semantic_types=datastructure.selected_cols_w_types,
            selected_cols=datastructure.selected_cols,
            schema=schema.get_table_schema(TABLE_NAME),
            data_split=DataSplit.TRAIN,
        )
        idx = 10
        assert isinstance(dataset[idx], tuple)
        assert len(dataset[idx]) == 2  # split into x,y
        assert len(dataset[idx][0]) == 2  # split into tabular, support
        assert len(dataset[idx][0][0]) == 2  # training cols  check
        assert len(dataset[idx][0][1]) == 2  # support cols check
        assert len([dataset[idx][1]]) == 1  # y check
        for i in range(0, len(dataset)):
            assert dataset[i][1] == 0  # all target values default to 0.

    def test_dataset_works_only_with_categorical_features(
        self, dataframe: pd.DataFrame
    ) -> None:
        """Test no errors are raised if the dataset only has categorical features."""
        datasource = DataFrameSource(dataframe.loc[:, ["M", "N", "TARGET"]])
        datasource.load_data()
        schema = BitfountSchema()
        schema.add_datasource_tables(
            datasource,
            force_stypes={TABLE_NAME: {"categorical": ["TARGET"]}},
            table_name=TABLE_NAME,
        )
        datastructure = DataStructure(target=["TARGET"], table=TABLE_NAME)
        datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
        dataset = _PyTorchDataset(
            datasource=datasource,
            target=["TARGET"],
            selected_cols_semantic_types=datastructure.selected_cols_w_types,
            selected_cols=datastructure.selected_cols,
            schema=schema.get_table_schema(TABLE_NAME),
            data_split=DataSplit.TRAIN,
        )
        assert "continuous" not in schema.tables[0].features
        idx = 10
        assert isinstance(dataset[idx], tuple)
        assert len(dataset[idx]) == 2  # split into x,y
        assert len(dataset[idx][0]) == 2  # split into tabular, support
        assert len(dataset[idx][0][0]) == 2  # training cols  check
        assert len(dataset[idx][0][1]) == 2  # support cols check
        assert len([dataset[idx][1]]) == 1  # y check

    def test_transform_image_with_default_batch_transformations(
        self, dataframe: pd.DataFrame
    ) -> None:
        """Test that the default transformations are applied to the image correctly.

        This is done for every split.
        """
        target = "TARGET"
        datasource = DataFrameSource(dataframe)
        datasource.load_data()
        schema = BitfountSchema(
            datasource,
            force_stypes={TABLE_NAME: {"image": ["image"]}},
            table_name=TABLE_NAME,
        )

        datasource.data = datasource.data.drop(
            columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
        )
        datastructure = DataStructure(
            target=target,
            table=TABLE_NAME,
            selected_cols=["image", target],
            image_cols=["image"],
        )
        datastructure.set_training_column_split_by_semantic_type(schema.tables[0])

        # Iterate over all data splits
        for split in DataSplit:
            image_dataset = _PyTorchDataset(
                datasource=datasource,
                target=target,
                schema=schema.get_table_schema(TABLE_NAME),
                selected_cols=datastructure.selected_cols,
                selected_cols_semantic_types=datastructure.selected_cols_w_types,
                batch_transforms=datastructure.get_batch_transformations(),
                data_split=split,
            )

            # Make assertions
            assert image_dataset.batch_transforms is not None
            img_array = np.array(Image.new("RGB", size=(224, 224), color=(55, 100, 2)))
            transformed_image = image_dataset._transform_image(img_array.copy(), 0)
            assert isinstance(transformed_image, torch.Tensor)

            # Torch transformation resizes the image so that the channels are first
            assert transformed_image.shape == (3, 224, 224)
            torch_img_array = torch.from_numpy(img_array)
            assert torch_img_array.shape == (224, 224, 3)

            # Check that the image dtypes are different because the transformed image
            # has been normalized. The actual tensors cannot be compared with eachother
            # because the dtypes are different.
            assert torch_img_array.dtype != transformed_image.dtype



"""Classes concerning data types."""
from __future__ import annotations

from abc import abstractmethod
import ast
from collections import Counter, OrderedDict
import copy
from enum import Enum
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
)

from marshmallow import Schema as MarshmallowSchema
from marshmallow import ValidationError, fields, post_dump, post_load, pre_dump
from marshmallow.fields import Field
from marshmallow_enum import EnumField
from mypy_extensions import Arg, KwArg
import numpy as np
import pandas as pd
from pandas._typing import Dtype
from pandas.core.dtypes.common import pandas_dtype
from typing_extensions import NotRequired

from bitfount.types import _Dtypes, _JSONDict
from bitfount.utils import delegates

if TYPE_CHECKING:
    from bitfount.data.datasources.base_source import BaseSource
    from bitfount.data.schema import TableSchema

logger = logging.getLogger(__name__)

# Type aliases for dunder methods
T = TypeVar("T")
_SingleOrMulti = Union[T, Sequence[T]]
_ImagesData = _SingleOrMulti[np.ndarray]
_TabularData = np.ndarray
_SupportData = np.ndarray
_Y_VAR = np.ndarray
_TabularEntry = Tuple[Tuple[_TabularData, _SupportData], _Y_VAR]
_ImageEntry = Tuple[Tuple[_ImagesData, _SupportData], _Y_VAR]
_ImageAndTabularEntry = Tuple[Tuple[_TabularData, _ImagesData, _SupportData], _Y_VAR]
_Segmentation_ImageEntry = Tuple[Tuple[_ImagesData, _SupportData], _ImagesData]
_Segmentation_ImageAndTabEntry = Tuple[
    Tuple[_TabularData, _ImagesData, _SupportData], _ImagesData
]
_DataEntry = Union[
    _TabularEntry,
    _ImageEntry,
    _ImageAndTabularEntry,
    _Segmentation_ImageEntry,
    _Segmentation_ImageAndTabEntry,
]
_DataBatch = List[_DataEntry]

# Type alias for the values in the `SemanticType` enum
_SemanticTypeValue = Literal["categorical", "continuous", "image", "text"]
S = TypeVar("S", bound="BaseSource")
_Column = Union[np.ndarray, pd.Series]
_GetColumnCallable = Callable[
    [Arg(S, "self"), Arg(str, "col_name"), KwArg(Any)], _Column  # noqa: F821
]
_GetDtypesCallable = Callable[[Arg(S, "self"), KwArg(Any)], _Dtypes]  # noqa: F821


class DataSplit(Enum):
    """Enum corresponding to the available data splits."""

    TRAIN = "train"
    VALIDATION = "validation"
    TEST = "test"


class _CamelCaseSchema(MarshmallowSchema):
    """Schema that uses camelCase for its external representation.

    snake_case is used for its internal representation.
    """

    @staticmethod
    def camel_case(s: str) -> str:
        """Converts a string from snake_case to camelCase."""
        parts = iter(s.split("_"))
        return next(parts) + "".join(i.title() for i in parts)

    def on_bind_field(self, field_name: str, field_obj: Field) -> None:
        """Converts the field's name to camelCase during hook."""
        field_obj.data_key = self.camel_case(field_obj.data_key or field_name)


class _ExtendableLabelEncoder:
    """Encodes strings as integers.

    A label encoder class which allows for building up the set of classes in multiple
    calls to `add_values` instead of having to create the set with one "fit" call.
    """

    def __init__(self) -> None:
        self.classes: Dict[str, int] = {}  # Mapping from class label to index
        self.dtype = str  # Fixed to str for now

    class _Schema(MarshmallowSchema):
        classes = fields.Dict(keys=fields.Str(), values=fields.Int())

        @post_load
        def recreate_encoder(
            self, data: _JSONDict, **_kwargs: Any
        ) -> _ExtendableLabelEncoder:
            """Recreates ExtendableLabelEncoder."""
            new_encoder = _ExtendableLabelEncoder()
            new_encoder.classes = dict(data["classes"])
            return new_encoder

    def add_values(self, values: Union[np.ndarray, pd.Series]) -> None:
        """Adds all entries in the column to the set."""
        uniques: Iterable
        if isinstance(values, np.ndarray):
            uniques = sorted(set(values.astype(self.dtype)))
        else:
            uniques = sorted(set(values.astype(self.dtype).tolist()))

        # Remove classes that are already present in `self.classes`
        new_uniques = sorted(u for u in uniques if u not in self.classes)
        cur_size = len(self.classes)

        # Adds a label for each new class by incrementing the previously largest label
        for i, new_val in enumerate(new_uniques):
            self.classes[new_val] = i + cur_size

    def add_values_with_encoding(self, values: Mapping[str, int]) -> None:
        """Adds all entries in the column to the set.

        This is used for the schema override in order to make
        sure that the encodings are the user defined encodings.

        Args:
            values: A mapping of strings to integer encodings, defined by the modeller.
        """
        for value, mapping in values.items():
            self.classes[value] = mapping

    def transform(self, values: pd.Series) -> List[int]:
        """Transforms the given column and returns it as a list of encoded values.

        Args:
            values: The column to encode as a `pandas.Series`.

        Returns:
            A list of encoded values.

        Raises:
            ValueError: If the encoder encounters a previously unseen label.
        """
        previously_unseen_label: str
        try:
            return [self.classes[v] for v in values.astype(self.dtype).tolist()]
        except KeyError as err:
            previously_unseen_label = str(err)
            logger.debug(f"Previously unseen label: {previously_unseen_label}")

        # If the encoder encounters a previously unseen label, we instead try to
        # interpret it as a datetime and see if there is a match there instead.
        # This is done because the same datetime can have multiple different formats
        # and pandas and numpy display these differently when converting to a string
        # so we try to convert them into the same format before encoding.
        logger.debug(f"Interpreting {previously_unseen_label} as a datetime.")
        try:
            classes = {}
            for _cls, val in self.classes.items():
                date = str(pd.to_datetime(_cls))
                classes[date] = val

            # We first try to see if there is a match once the schema label has been
            # converted into a datetime. If there is no match, this could be due to an
            # empty time component in the value to be encoded so we try again after
            # adding this.
            return [
                classes.get(v, classes[str(pd.to_datetime(v))])
                for v in values.astype(self.dtype).tolist()
            ]
        except Exception:
            logger.debug(
                f"Failed to interpret {previously_unseen_label} as a datetime."
            )

        raise ValueError(f"Previously unseen label: {previously_unseen_label}")

    @property
    def size(self) -> int:
        """Number of values in the encoder."""
        return len(self.classes)

    def __eq__(self, other: Any) -> bool:
        if self.classes == other.classes and self.dtype == other.dtype:
            return True
        return False

    def __hash__(self) -> int:
        return hash((self.classes, self.dtype))


class SemanticType(Enum):
    """Simple wrapper for some basic data types."""

    CATEGORICAL = "categorical"
    CONTINUOUS = "continuous"
    IMAGE = "image"
    TEXT = "text"


class _SemanticTypeRecord:
    """Simple semantic type wrapper for an individual record.

    Args:
        feature_name (str): name of the feature
        dtype (Union[Dtype, np.dtype]): data type of the feature
        description (str, optional): description of the feature. Defaults to None.
    """

    def __init__(
        self,
        feature_name: str,
        dtype: Union[Dtype, np.dtype],
        description: Optional[str] = None,
    ) -> None:
        self.feature_name = feature_name
        self.dtype = dtype
        self.description = description

    @property
    @abstractmethod
    def semantic_type(self) -> SemanticType:
        """Returns the relevant SemanticType for the class."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def add_record_to_schema(
        cls, schema: TableSchema, **constructor_arguments: Any
    ) -> None:
        """Creates record and adds it to schema features."""
        raise NotImplementedError

    class _Schema(_CamelCaseSchema):

        feature_name = fields.Str()
        dtype = fields.Str()
        semantic_type = EnumField(
            SemanticType,
            by_value=True,
        )
        description = fields.Str(allow_none=True)

        @post_dump
        def sort_alphabetically(self, data: _JSONDict, **kwargs: Any) -> _JSONDict:
            """Sorts the keys of the dictionary alphabetically after dumping.

            The exception is `featureName` which is moved to be the first key.
            """
            data = OrderedDict(dict(sorted(data.items())))
            data.move_to_end("featureName", last=False)
            return dict(data)

        @staticmethod
        def convert_dtype(data: _JSONDict) -> _JSONDict:
            """Converts `dtype` from string representation to actual dtype.

            Raises:
                ValidationError: if the dtype can't be deciphered
            """
            try:
                data["dtype"] = pandas_dtype(data["dtype"])
            except TypeError:
                raise ValidationError(
                    f"Continuous record `dtype` expected a valid np.dtype or a "
                    f"pandas dtype but received: `{data['dtype']}`."
                )
            return data


@delegates()
class CategoricalRecord(_SemanticTypeRecord):
    """Stores information for a categorical feature in the schema.

    Args:
        encoder: An encoder for the different categories. Defaults to None.
    """

    def __init__(
        self,
        encoder: Optional[_ExtendableLabelEncoder] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        self.encoder: _ExtendableLabelEncoder = (
            encoder if encoder is not None else _ExtendableLabelEncoder()
        )

    @property
    def semantic_type(self) -> SemanticType:
        """Property for the relevant `SemanticType` for the class.

        Returns:
            The categorical `SemanticType`.
        """
        return SemanticType.CATEGORICAL

    @classmethod
    def add_record_to_schema(
        cls, schema: TableSchema, **constructor_arguments: Any
    ) -> None:
        """Create a categorical record and add it to the schema features.

        Args:
            schema: A `TableSchema` object.
            **constructor_arguments: Additional arguments to pass to the schema class.
        """
        record = cls(**constructor_arguments)
        if "categorical" not in schema.features:
            schema.features["categorical"] = {record.feature_name: record}
        else:
            schema.features["categorical"][record.feature_name] = record

    class _Schema(_SemanticTypeRecord._Schema):

        encoder = fields.Nested(_ExtendableLabelEncoder._Schema)

        @post_load
        def recreate_record(self, data: _JSONDict, **_kwargs: Any) -> CategoricalRecord:
            """Recreates CategoricalRecord."""
            data = self.convert_dtype(data)
            return CategoricalRecord(**data)


@delegates()
class TextRecord(_SemanticTypeRecord):
    """Stores information for a text feature in the schema."""

    @property
    def semantic_type(self) -> SemanticType:
        """Property for the relevant `SemanticType` for the class.

        Returns:
            The text `SemanticType`.
        """
        return SemanticType.TEXT

    @classmethod
    def add_record_to_schema(
        cls, schema: TableSchema, **constructor_arguments: Any
    ) -> None:
        """Create a text record and add it to schema features.

        Args:
            schema: A `TableSchema` object.
            **constructor_arguments: Additional arguments to pass to the schema class.
        """
        record = cls(**constructor_arguments)
        if "text" not in schema.features:
            schema.features["text"] = {record.feature_name: record}
        else:
            schema.features["text"][record.feature_name] = record

    class _Schema(_SemanticTypeRecord._Schema):
        @post_load
        def recreate_record(self, data: _JSONDict, **_kwargs: Any) -> TextRecord:
            """Recreates TextRecord."""
            data = self.convert_dtype(data)
            return TextRecord(**data)


@delegates()
class ContinuousRecord(_SemanticTypeRecord):
    """Stores information for a continuous feature in the schema."""

    @property
    def semantic_type(self) -> SemanticType:
        """Property for the relevant `SemanticType` for the class.

        Returns:
            The continuous `SemanticType`.
        """
        return SemanticType.CONTINUOUS

    @classmethod
    def add_record_to_schema(
        cls, schema: TableSchema, **constructor_arguments: Any
    ) -> None:
        """Create a continuous record and add it to schema features.

        Args:
            schema: A `TableSchema` object.
            **constructor_arguments: Additional arguments to pass to the schema class.
        """
        record = cls(**constructor_arguments)
        if "continuous" not in schema.features:
            schema.features["continuous"] = {record.feature_name: record}
        else:
            schema.features["continuous"][record.feature_name] = record

    class _Schema(_SemanticTypeRecord._Schema):
        @post_load
        def recreate_record(self, data: _JSONDict, **_kwargs: Any) -> ContinuousRecord:
            """Recreates ContinuousRecord."""
            data = self.convert_dtype(data)
            return ContinuousRecord(**data)


class StrDictField(fields.Field):
    """Field that can be str or dict."""

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs: Any,
    ) -> Any:
        if isinstance(value, str) or isinstance(value, dict):
            return value
        else:
            raise ValidationError("Field should be str or dict")


@delegates()
class ImageRecord(_SemanticTypeRecord):
    """Stores information for an image feature in the schema."""

    def __init__(
        self,
        dimensions: Optional[Counter] = None,
        modes: Optional[Counter] = None,
        formats: Optional[Counter] = None,
        **kwargs: Any,
    ) -> None:
        """Stores information for an image feature in the schema.

        Args:
            dimensions: The dimensions of the different images in
                the column. Defaults to None.
            modes: The modes of the different images in the column.
                Defaults to None.
            formats: The formats of the different images in the
                column. Defaults to None.
        """
        super().__init__(**kwargs)

        self.dimensions: Counter = dimensions if dimensions is not None else Counter()
        self.modes: Counter = modes if modes is not None else Counter()
        self.formats: Counter = formats if formats is not None else Counter()

    @property
    def semantic_type(self) -> SemanticType:
        """Property for the relevant `SemanticType` for the class.

        Returns:
            The image `SemanticType`.
        """
        return SemanticType.IMAGE

    @classmethod
    def add_record_to_schema(
        cls, schema: TableSchema, **constructor_arguments: Any
    ) -> None:
        """Create an image record and add it to schema features.

        Args:
            schema: A `TableSchema` object.
            **constructor_arguments: Additional arguments to pass to the schema class.
        """
        record = cls(**constructor_arguments)
        if "image" not in schema.features:
            schema.features["image"] = {record.feature_name: record}
        else:
            schema.features["image"][record.feature_name] = record

    class _Schema(_SemanticTypeRecord._Schema):

        dimensions = fields.Dict()
        formats = fields.Dict()
        modes = fields.Dict()

        @pre_dump
        def get_image_features(self, obj: ImageRecord, **_kwargs: Any) -> ImageRecord:
            """Converts image features from Counters to dictionaries."""
            temp_schema = copy.deepcopy(obj)
            # Ignoring these mypy assignment errors so that we can dump the image
            # properties as dictionaries for ease and readability
            temp_schema.dimensions = {  # type: ignore[assignment] # Reason: see comment
                str(key): value
                for key, value in dict(obj.dimensions.most_common()).items()
            }
            temp_schema.modes = {  # type: ignore[assignment] # Reason: see comment
                str(key): value for key, value in dict(obj.modes.most_common()).items()
            }
            temp_schema.formats = {  # type: ignore[assignment] # Reason: see comment
                str(key): value
                for key, value in dict(obj.formats.most_common()).items()
            }
            return temp_schema

        @post_load
        def deserialize_image_features(
            self, data: _JSONDict, **_kwargs: Any
        ) -> ImageRecord:
            """Converts image features back to Counters from dictionaries."""
            data["dimensions"] = Counter(
                {
                    ast.literal_eval(key): value
                    for key, value in data["dimensions"].items()
                }
            )
            data["modes"] = Counter(data["modes"])
            data["formats"] = Counter(data["formats"])
            data = self.convert_dtype(data)
            return ImageRecord(**data)


class _FeatureDict(TypedDict):
    """Typed dictionary for the features in a TableSchema.

    NotRequired indicates that the keys don't all need to be present. But the keys
    that are present need to be one of the ones listed below.
    """

    categorical: NotRequired[Dict[str, CategoricalRecord]]
    continuous: NotRequired[Dict[str, ContinuousRecord]]
    image: NotRequired[Dict[str, ImageRecord]]
    text: NotRequired[Dict[str, TextRecord]]


class DataPathModifiers(TypedDict):
    """TypedDict class for path modifiers.

    NotRequired indicates that the keys don't all need to be present. But the keys
    that are present need to be one of the ones listed below.

    Args:
        suffix: The suffix to be used for modifying a path.
        prefix: The prefix to be used for modifying a path.
    """

    suffix: NotRequired[str]
    prefix: NotRequired[str]


SchemaOverrideMapping = Mapping[
    _SemanticTypeValue, List[Union[str, Mapping[str, Mapping[str, int]]]]
]



"""Test dataset classes in data/datasets.py."""
import logging
from typing import Iterator, Optional, Tuple
from unittest.mock import Mock

from PIL import Image
from _pytest.logging import LogCaptureFixture
import numpy as np
import pandas as pd
import pytest
from pytest import fixture
from pytest_mock import MockerFixture
import sqlalchemy

from bitfount.data.datasets import _BitfountDataset, _IterableBitfountDataset
from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.datasplitters import (
    DatasetSplitter,
    PercentageSplitter,
    SplitterDefinedInData,
)
from bitfount.data.datastructure import DataStructure
from bitfount.data.exceptions import DataNotLoadedError
from bitfount.data.schema import BitfountSchema
from bitfount.data.types import DataSplit, SemanticType
from bitfount.data.utils import DatabaseConnection, _generate_dtypes_hash
from tests.utils.helper import (
    DATASET_ROW_COUNT,
    TABLE_NAME,
    create_dataset,
    integration_test,
    unit_test,
)


@fixture
def tabular_dataframe() -> pd.DataFrame:
    """Underlying dataframe for tabular datasets."""
    return create_dataset()


class FakeSplitter(DatasetSplitter):
    """Fake Splitter that just returns predefined indices."""

    def __init__(
        self,
        train_indices: np.ndarray,
        validation_indices: np.ndarray,
        test_indices: np.ndarray,
    ):
        self.train_indices = train_indices
        self.validation_indices = validation_indices
        self.test_indices = test_indices

    @classmethod
    def splitter_name(cls) -> str:
        """Splitter name for config."""
        return "FakeSplitter"

    def create_dataset_splits(
        self, data: pd.DataFrame
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Returns predefined indices and provided data."""
        return self.train_indices, self.validation_indices, self.test_indices

    def get_split_query(self, datasource: DatabaseSource, split: DataSplit) -> str:
        """Returns query for given split."""


class TestBaseBitfountDataset:
    """Tests BaseBitfountDataset class."""

    @fixture
    def test_percentage(self) -> int:
        """Test percentage value."""
        return 25

    @fixture
    def validation_percentage(self) -> int:
        """Validation percentage value."""
        return 55

    @fixture
    def train_percentage(self, test_percentage: int, validation_percentage: int) -> int:
        """Train percentage value."""
        return (100 - test_percentage) - validation_percentage

    @fixture
    def target(self) -> str:
        """Target value."""
        return "TARGET"

    @fixture
    def fake_splitter(
        self, test_percentage: int, train_percentage: int, validation_percentage: int
    ) -> FakeSplitter:
        """Fake splitter defined using train/validation/test values."""
        return FakeSplitter(
            np.array(range(int(DATASET_ROW_COUNT * (train_percentage / 100)))),
            np.array(range(int(DATASET_ROW_COUNT * (validation_percentage / 100)))),
            np.array(range(int(DATASET_ROW_COUNT * (test_percentage / 100)))),
        )

    @fixture
    def datastructure(self, target: str) -> DataStructure:
        """Datastructure fixture."""
        return DataStructure(target=target, ignore_cols=["image"], table=TABLE_NAME)

    @unit_test
    def test_transform_image_with_custom_batch_transformation(
        self, image_dataset: _BitfountDataset
    ) -> None:
        """Test transform_image method."""
        assert image_dataset.batch_transforms is not None
        img_array = np.array(Image.new("RGB", size=(224, 224), color=(55, 100, 2)))
        transformed_image = image_dataset._transform_image(img_array.copy(), 0)
        assert isinstance(transformed_image, np.ndarray)
        assert transformed_image.shape == (224, 224, 3)

        # Assert that the transformed image is not the same as the original
        with pytest.raises(AssertionError):
            np.testing.assert_array_equal(img_array, transformed_image)

    @unit_test
    def test_load_image(self, image_dataset: _BitfountDataset) -> None:
        """Test _load_image method."""
        loaded_transformed_image = image_dataset._load_images(0)
        assert isinstance(loaded_transformed_image, np.ndarray)
        assert loaded_transformed_image.shape == (224, 224, 3)

    @unit_test
    def test_load_target_image(self, seg_image_dataset: _BitfountDataset) -> None:
        """Test _load_image method for target image."""
        loaded_transformed_image = seg_image_dataset._load_images(
            0, what_to_load="target"
        )
        assert isinstance(loaded_transformed_image, np.ndarray)
        assert loaded_transformed_image.shape == (100, 100, 3)

    @unit_test
    def test_apply_schema(
        self, caplog: LogCaptureFixture, tabular_dataset: pd.DataFrame
    ) -> None:
        """Tests data is loaded according to `selected_cols` and schema."""
        assert set(tabular_dataset.data.columns).issubset(
            set(tabular_dataset.selected_cols)
        )
        assert set(tabular_dataset.data.columns).issubset(
            set(tabular_dataset.schema.get_feature_names())
        )

        for record in caplog.records:
            if record.levelname == "WARNING":
                assert (
                    record.message
                    == "Selected columns `I,J,K,L` were not found in the data, continuing without them."  # noqa: B950
                )

    @unit_test
    def test_apply_schema_with_extra_columns(
        self,
        caplog: LogCaptureFixture,
        datastructure: DataStructure,
        tabular_dataframe: pd.DataFrame,
    ) -> None:
        """Tests data is loaded according to `selected_cols` and schema.

        Checks that even if extra columns are provided in the `selected_cols` from the
        DataStructure, they are not used when loading the datasource.
        """
        datasource = DataFrameSource(tabular_dataframe, ignore_cols=["image"])
        datasource.load_data()

        schema = BitfountSchema()
        schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
        datasource.data = datasource.data.drop(
            columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
        )
        datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
        tabular_dataset = _BitfountDataset(
            datasource=datasource,
            selected_cols=datastructure.selected_cols,
            data_split=DataSplit.TRAIN,
            schema=schema.get_table_schema(TABLE_NAME),
            selected_cols_semantic_types=datastructure.selected_cols_w_types,
        )
        assert tabular_dataset.datasource.data is not None

        assert set(tabular_dataset.datasource.data.columns).issubset(
            set(tabular_dataset.selected_cols)
        )
        assert set(tabular_dataset.datasource.data.columns).issubset(
            set(tabular_dataset.schema.get_feature_names())
        )
        assert (
            "Selected columns `I,J,K,L` were not found in the data, continuing without them."  # noqa: B950
            in caplog.text
        )

    @unit_test
    def test_datasource_data_split_called_twice(
        self, caplog: LogCaptureFixture, iter_tabular_dataset: _IterableBitfountDataset
    ) -> None:
        """Tests that the log is printed if split_data called twice."""
        caplog.set_level(logging.DEBUG)

        assert iter_tabular_dataset.datasource._data_is_split is False

        iter_tabular_dataset._split_data()

        assert iter_tabular_dataset.datasource._data_is_split is True
        # split data in DataBunch into training, validation, test sets
        iter_tabular_dataset._split_data()
        assert "Data is already split, keeping the current split." in caplog.text

    @pytest.mark.parametrize(
        "datasplit_type", [DataSplit.TEST, DataSplit.TRAIN, DataSplit.VALIDATION]
    )
    @unit_test
    def test_X_set(
        self,
        dataframe: pd.DataFrame,
        datasplit_type: DataSplit,
        datastructure: DataStructure,
        fake_splitter: FakeSplitter,
        target: str,
        test_percentage: int,
        train_percentage: int,
        validation_percentage: int,
    ) -> None:
        """Checks test/train/validation sets behave correctly."""
        if datasplit_type == DataSplit.TEST:
            datasplit_val = test_percentage
        elif datasplit_type == DataSplit.TRAIN:
            datasplit_val = train_percentage
        elif datasplit_type == DataSplit.VALIDATION:
            datasplit_val = validation_percentage

        datasource = DataFrameSource(dataframe, data_splitter=fake_splitter)
        datasource.load_data()

        schema = BitfountSchema()
        schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
        datasource.data = datasource.data.drop(
            columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
        )

        datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
        dataset = _BitfountDataset(
            datasource=datasource,
            target=target,
            selected_cols=datastructure.selected_cols,
            data_split=DataSplit.TRAIN,
            schema=schema.get_table_schema(TABLE_NAME),
            selected_cols_semantic_types=datastructure.selected_cols_w_types,
        )

        dataset._split_data()
        assert dataset.datasource.data is not None
        # assert columns match the original data
        assert (
            dataset.datasource.data.shape[1]
            == dataset.get_dataset_split(datasplit_type).shape[1]
        )
        # assert there are the expected number of rows
        assert (
            int(datasplit_val * dataset.datasource.data.shape[0] / 100)
            == dataset.get_dataset_split(datasplit_type).shape[0]
        )

    @unit_test
    @pytest.mark.parametrize(
        "datasource_splitter, datastructure_splitter",
        [
            (PercentageSplitter(), PercentageSplitter()),
            (PercentageSplitter(), None),
            (None, PercentageSplitter()),
            (None, None),
        ],
    )
    def test_resolve_data_splitter(
        self,
        dataframe: pd.DataFrame,
        datasource_splitter: Optional[DatasetSplitter],
        datastructure_splitter: Optional[DatasetSplitter],
    ) -> None:
        """Checks data splitter is resolved correctly.

        If datasource has a data_splitter, use it to split the data.
        Else if the datastructure has a data_splitter use that splitter.
        Else use the PercentageSplitter.
        """
        datasource = DataFrameSource(dataframe, data_splitter=datasource_splitter)
        datasource.load_data()
        target = "TARGET"

        schema = BitfountSchema()
        schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
        datasource.data = datasource.data.drop(
            columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
        )
        datastructure = DataStructure(
            target=target, ignore_cols=["image"], table=TABLE_NAME
        )
        datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
        dataset = _IterableBitfountDataset(
            datasource=datasource,
            target=target,
            selected_cols_semantic_types=datastructure.selected_cols_w_types,
            selected_cols=datastructure.selected_cols,
            data_split=DataSplit.TRAIN,
            schema=schema.get_table_schema(TABLE_NAME),
        )
        resolved_splitter = dataset._resolve_data_splitter(datastructure_splitter)
        assert isinstance(resolved_splitter, PercentageSplitter)

        if datasource_splitter:
            assert datasource.data_splitter is not None
            assert resolved_splitter is datasource.data_splitter
            assert resolved_splitter is not datastructure_splitter
        elif datastructure_splitter:
            assert datasource.data_splitter is None
            assert resolved_splitter is datastructure_splitter
        else:
            assert resolved_splitter is not datastructure_splitter
            assert resolved_splitter is not datasource_splitter

    @unit_test
    def test_zero_validation_test_size(
        self, dataframe: pd.DataFrame, target: str
    ) -> None:
        """Checks Dataset object behaves properly when if valid and test pct are 0."""
        datasource = DataFrameSource(
            dataframe,
            data_splitter=FakeSplitter(
                train_indices=np.array(range(DATASET_ROW_COUNT)),
                validation_indices=np.array([]),
                test_indices=np.array([]),
            ),
        )
        datasource.load_data()
        schema = BitfountSchema()
        schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
        datasource.data = datasource.data.drop(
            columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
        )
        datastructure = DataStructure(
            target=target, ignore_cols=["image"], table=TABLE_NAME
        )
        datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
        dataset = _BitfountDataset(
            datasource=datasource,
            target=target,
            selected_cols=datastructure.selected_cols,
            data_split=DataSplit.TRAIN,
            schema=schema.get_table_schema(TABLE_NAME),
            selected_cols_semantic_types=datastructure.selected_cols_w_types,
        )
        assert len(datasource.data) == len(dataset.get_dataset_split(DataSplit.TRAIN))
        assert datasource._test_idxs is not None
        assert datasource._validation_idxs is not None
        assert len(datasource._test_idxs) == 0
        assert len(datasource._validation_idxs) == 0

    @integration_test
    def test_mock_database_query_input(
        self, db_session: sqlalchemy.engine.base.Engine, mocker: MockerFixture
    ) -> None:
        """Checks DatabaseSource initialises correctly with `DatabaseConnection`."""
        db_conn = DatabaseConnection(
            db_session,
            query="""
            SELECT 'd1.Date', 'A'
            FROM dummy_data d1
            LEFT JOIN dummy_data_2 d2
            ON 'd1.Date' = 'd2.Date'
            """,
        )
        datasource = DatabaseSource(db_conn, seed=420)
        datasource.load_data()

        mocker.patch.object(_BitfountDataset, "_set_column_name_attributes")
        mocker.patch.object(_BitfountDataset, "_reformat_data")

        db_dataset = _BitfountDataset(
            datasource=datasource,
            data_split=Mock(value="train"),
            schema=Mock(),
            selected_cols=Mock(),
            selected_cols_semantic_types=Mock(),
        )
        db_dataset._split_data()
        assert not datasource.multi_table
        assert datasource.data is not None
        assert datasource._train_idxs is not None
        assert datasource._validation_idxs is not None
        assert datasource._test_idxs is not None
        assert len(datasource._train_idxs) + len(datasource._validation_idxs) + len(
            datasource._test_idxs
        ) == len(datasource.data)

    @unit_test
    def test_training_set_unloaded_raises_error(
        self, mock_engine: Mock, mocker: MockerFixture
    ) -> None:
        """Tests that calling the `training_set` property raises an error.

        The error should be a `DataNotLoadedError` when the data is multi-table because
        it is never loaded.
        """
        target = "TARGET"
        table_name = "dummy_data"
        datastructure = DataStructure(
            target=target, ignore_cols=["image"], table=table_name
        )

        db_conn = DatabaseConnection(
            mock_engine, table_names=["dummy_data", "dummy_data_2"]
        )
        datasource = DatabaseSource(db_conn, seed=420)
        mock_dtypes = {"col": np.dtype(int)}
        mocker.patch.object(datasource, "get_dtypes", return_value=mock_dtypes)
        datasource._table_hashes.add(_generate_dtypes_hash(mock_dtypes))
        datasource.load_data()
        schema = BitfountSchema()
        schema.add_datasource_tables(datasource, table_name=table_name)
        # Using IterableBitfountDataset but get_dataset_split is a
        # concrete method of _BaseBitfountDataset
        db_dataset = _IterableBitfountDataset(
            datasource=datasource,
            target=target,
            selected_cols=datastructure.selected_cols,
            data_split=DataSplit.TRAIN,
            schema=schema.get_table_schema(table_name),
            selected_cols_semantic_types=datastructure.selected_cols_w_types,
        )
        with pytest.raises(DataNotLoadedError):
            db_dataset.get_dataset_split(DataSplit.TRAIN)

    @unit_test
    def test_datasource_data_split_flag_updated(self, dataframe: pd.DataFrame) -> None:
        """Tests that the data_is_split flag is updated."""
        target = "TARGET"
        ds = DataFrameSource(dataframe, selected_cols=["M", "F"])
        ds.load_data()
        assert ds._data_is_split is False

        schema = BitfountSchema()
        schema.add_datasource_tables(ds, table_name=TABLE_NAME)
        datastructure = DataStructure(
            target=target, ignore_cols=["image"], table=TABLE_NAME
        )
        datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
        dataset = _IterableBitfountDataset(
            datasource=ds,
            selected_cols_semantic_types=datastructure.selected_cols_w_types,
            selected_cols=datastructure.selected_cols,
            data_split=DataSplit.TRAIN,
            schema=schema.get_table_schema(TABLE_NAME),
        )
        dataset._split_data()
        assert ds._data_is_split is True

    @unit_test
    def test_get_dataset_split_length_returns_length_correctly(
        self, dataframe: pd.DataFrame, target: str
    ) -> None:
        """Tests that `get_dataset_split_length` returns the correct length.

        This is done with the `SplitterDefinedInData` splitter.
        """
        dataframe["BITFOUNT_SPLIT_CATEGORY"] = np.random.choice(
            ["TRAIN", "VALIDATE", "TEST"], size=dataframe.shape[0]
        )

        ds = DataFrameSource(dataframe, data_splitter=SplitterDefinedInData())
        ds.load_data()
        schema = BitfountSchema()
        schema.add_datasource_tables(ds, table_name=TABLE_NAME)
        datastructure = DataStructure(
            target=target, ignore_cols=["image"], table=TABLE_NAME
        )
        datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
        dataset = _BitfountDataset(
            datasource=ds,
            target=target,
            selected_cols_semantic_types=datastructure.selected_cols_w_types,
            selected_cols=datastructure.selected_cols,
            data_split=DataSplit.TRAIN,
            schema=schema.get_table_schema(TABLE_NAME),
        )
        train_length = dataset.get_dataset_split_length(DataSplit.TRAIN)
        assert train_length == len(
            dataframe.loc[dataframe["BITFOUNT_SPLIT_CATEGORY"] == "TRAIN"]
        )

    @unit_test
    def test_get_dataset_split_length_executes_query(
        self,
        mock_engine: Mock,
        mocker: MockerFixture,
    ) -> None:
        """Tests that `get_dataset_split_length` executes query with database."""
        db_conn = DatabaseConnection(
            mock_engine, table_names=["dummy_data", "dummy_data_2"]
        )
        datasource = DatabaseSource(
            db_conn,
            data_splitter=SplitterDefinedInData(
                column_name='q."M"', training_set_label="True"
            ),
        )
        datasource.load_data(sql_query="SELECT * FROM dummy_data")
        mock_connect = mocker.patch.object(datasource.con, "connect")
        mock_connect.__enter__.return_value = Mock()
        mocker.patch.object(_IterableBitfountDataset, "_set_column_name_attributes")
        dataset = _IterableBitfountDataset(datasource, Mock(), Mock(), Mock(), Mock())
        dataset.get_dataset_split_length(DataSplit.TRAIN)

        mock_connect.assert_called_once()

    @integration_test
    def test_get_dataset_split_length_with_multitable_database_connection(
        self,
        dataframe: pd.DataFrame,
        db_session: sqlalchemy.engine.base.Engine,
        mocker: MockerFixture,
    ) -> None:
        """Tests that `get_dataset_split_length` returns the correct length."""
        db_conn = DatabaseConnection(
            db_session, table_names=["dummy_data", "dummy_data_2"]
        )
        datasource = DatabaseSource(
            db_conn,
            data_splitter=SplitterDefinedInData(
                column_name='q."M"', training_set_label="True"
            ),
        )
        datasource.load_data(sql_query="SELECT * FROM dummy_data")
        mocker.patch.object(_IterableBitfountDataset, "_set_column_name_attributes")
        dataset = _IterableBitfountDataset(datasource, Mock(), Mock(), Mock(), Mock())
        train_length = dataset.get_dataset_split_length(DataSplit.TRAIN)
        assert train_length == len(dataframe.loc[dataframe["M"] == True])  # noqa: E712

    @unit_test
    def test_unable_get_dataset_split_length(
        self, dataframe: pd.DataFrame, fake_splitter: FakeSplitter, target: str
    ) -> None:
        """Test error raise when unable to split dataset."""
        ds = DataFrameSource(dataframe, data_splitter=fake_splitter)
        ds.load_data()

        schema = BitfountSchema()
        schema.add_datasource_tables(ds, table_name=TABLE_NAME)
        datastructure = DataStructure(
            target=target, ignore_cols=["image"], table=TABLE_NAME
        )
        datastructure.set_training_column_split_by_semantic_type(schema.tables[0])

        dataset = _BitfountDataset(
            datasource=ds,
            target=target,
            selected_cols_semantic_types=datastructure.selected_cols_w_types,
            selected_cols=datastructure.selected_cols,
            data_split=DataSplit.TRAIN,
            schema=schema.get_table_schema(TABLE_NAME),
        )
        dataset.datasource._data_is_loaded = False
        with pytest.raises(
            DataNotLoadedError,
            match="Unable to get length of dataset split",
        ):
            dataset.get_dataset_split_length(DataSplit.TRAIN)


@unit_test
class TestBitfountDataset:
    """Tests BitfountDataset class."""

    def test_len_tab_data(self, tabular_dataset: _BitfountDataset) -> None:
        """Tests tabular dataset __len__ method."""
        assert tabular_dataset.datasource._train_idxs is not None
        assert len(tabular_dataset) == len(tabular_dataset.datasource._train_idxs)

    def test_len_img_data(self, image_tab_dataset: _BitfountDataset) -> None:
        """Tests image dataset __len__ method."""
        assert image_tab_dataset.datasource._train_idxs is not None
        assert len(image_tab_dataset) == len(image_tab_dataset.datasource._train_idxs)

    def test_len_img_tab_data(self, image_dataset: _BitfountDataset) -> None:
        """Tests dataset __len__ method."""
        assert image_dataset.datasource._train_idxs is not None
        assert len(image_dataset) == len(image_dataset.datasource._train_idxs)

    def test_len_multiimg_data(self, multiimage_dataset: _BitfountDataset) -> None:
        """Tests multi-image dataset __len__ method."""
        assert multiimage_dataset.datasource._train_idxs is not None
        assert len(multiimage_dataset) == len(multiimage_dataset.datasource._train_idxs)


class TestIterableBitfountDataset:
    """Tests IterableBitfountDataset class."""

    @pytest.mark.parametrize(
        "data_split", [DataSplit.TRAIN, DataSplit.VALIDATION, DataSplit.TEST]
    )
    @pytest.mark.parametrize(
        "validation_percentage,test_percentage", [(0, 0), (10, 10)]
    )
    @unit_test
    def test_len_magic_method(
        self,
        data_split: DataSplit,
        mock_engine: Mock,
        mocker: MockerFixture,
        tabular_dataframe: pd.DataFrame,
        test_percentage: int,
        validation_percentage: int,
    ) -> None:
        """Tests that __len__ magic method returns correct row count."""
        db_conn = DatabaseConnection(
            mock_engine,
            table_names=["dummy_data", "dummy_data_2"],
        )
        splitter = PercentageSplitter(
            validation_percentage=validation_percentage,
            test_percentage=test_percentage,
        )
        ds = DatabaseSource(
            db_conn,
            seed=420,
            data_splitter=splitter,
        )
        mock_len = mocker.patch.object(
            DatabaseSource, "__len__", return_value=DATASET_ROW_COUNT
        )

        schema = BitfountSchema(
            datasource=DataFrameSource(tabular_dataframe), table_name="dummy_data"
        )
        dataset = _IterableBitfountDataset(
            datasource=ds,
            selected_cols_semantic_types={
                "continuous": ["columns", "dont", "matter", "here"]
            },
            selected_cols=[],
            schema=schema.tables[0],
            target="TARGET",
            data_split=data_split,
        )

        # Call __len__ method on dataset
        dataset_length = len(dataset)

        mock_len.assert_called_once()

        if data_split == DataSplit.TRAIN:
            assert dataset_length == int(
                DATASET_ROW_COUNT * splitter.train_percentage / 100
            )
        elif data_split == DataSplit.VALIDATION:
            assert dataset_length == int(
                DATASET_ROW_COUNT * splitter.validation_percentage / 100
            )
        elif data_split == DataSplit.TEST:
            assert dataset_length == int(
                DATASET_ROW_COUNT * splitter.test_percentage / 100
            )

    @unit_test
    def test_dataset_len_is_cached(
        self, mock_engine: Mock, mocker: MockerFixture, tabular_dataframe: pd.DataFrame
    ) -> None:
        """Tests that __len__ magic method uses cached length."""
        db_conn = DatabaseConnection(
            mock_engine,
            table_names=["dummy_data", "dummy_data_2"],
        )
        splitter = PercentageSplitter()  # default is 80:10:10
        ds = DatabaseSource(
            db_conn,
            seed=420,
            data_splitter=splitter,
        )
        mock_len = mocker.patch.object(
            DatabaseSource, "__len__", return_value=DATASET_ROW_COUNT
        )

        schema = BitfountSchema(
            datasource=DataFrameSource(tabular_dataframe), table_name="dummy_data"
        )
        dataset = _IterableBitfountDataset(
            datasource=ds,
            selected_cols_semantic_types={
                "continuous": ["columns", "dont", "matter", "here"]
            },
            selected_cols=[],
            schema=schema.tables[0],
            target="TARGET",
            data_split=DataSplit.TRAIN,
        )

        # Call __len__ method on dataset TWICE
        dataset_length = len(dataset)
        dataset_length_2 = len(dataset)

        # Make assertion that mock only called ONCE
        mock_len.assert_called_once()

        # Makes assertion on final result
        assert (
            dataset_length
            == dataset_length_2
            == int(DATASET_ROW_COUNT * splitter.train_percentage / 100)
        )

    @unit_test
    def test_iter_magic_method(
        self, mock_engine: Mock, mocker: MockerFixture, tabular_dataframe: pd.DataFrame
    ) -> None:
        """Tests that __iter__ magic method works as expected."""
        ds = DataFrameSource(tabular_dataframe)
        schema = BitfountSchema(
            datasource=ds,
            table_name="dummy_data",
            force_stypes={"dummy_data": {"categorical": ["TARGET"]}},
        )

        class MockBaseSourceIterator:
            """Mock class to represent database result paritions."""

            def __iter__(self) -> Iterator[pd.DataFrame]:
                """Iterator just returns one set of dataframe values."""
                yield tabular_dataframe

        dataset = _IterableBitfountDataset(
            datasource=ds,
            selected_cols_semantic_types={
                "continuous": ["A", "B"],
                "categorical": ["TARGET"],
            },
            selected_cols=["TARGET", "A", "B"],
            schema=schema.tables[0],
            target="TARGET",
            data_split=DataSplit.TRAIN,
        )

        mock_datasource_iterator = MockBaseSourceIterator()
        mocker.patch.object(
            dataset, "yield_dataset_split", return_value=mock_datasource_iterator
        )

        # Call __iter__ method on dataset
        dataset_iterator = iter(dataset)
        row = next(dataset_iterator)  # First output of iterator
        assert isinstance(row, tuple)
        assert isinstance(row[0], tuple)  # x data
        assert isinstance(row[0][0], np.ndarray)  # tabular x data
        assert isinstance(row[0][1], np.ndarray)  # x support columns
        assert isinstance(row[1], np.integer)  # y data

        dataset.yield_dataset_split.assert_called_once()  # type: ignore[attr-defined] # Reason: check if mocked method called. # noqa: B950

    @unit_test
    def test_yield_dataset_split_raises_value_error_if_not_database_loader(
        self, iter_tabular_dataset: _IterableBitfountDataset
    ) -> None:
        """Tests that ValueError is raised if loader is not a DatabaseSource."""
        with pytest.raises(
            ValueError,
            match="Iterating over in-memory data is currently not supported.",
        ):
            next(iter_tabular_dataset.yield_dataset_split(DataSplit.TRAIN))

    @integration_test
    def test_yield_dataset_split_raises_value_error_if_no_query_or_table_name_provided(
        self,
        iter_db_dataset: _IterableBitfountDataset,
    ) -> None:
        """Tests that ValueError is raised if no query or table name provided."""
        with pytest.raises(ValueError, match="No query or table name specified."):
            next(iter_db_dataset.yield_dataset_split(DataSplit.TRAIN))

    @integration_test
    def test_validation_set_unsplit_raises_error(
        self, iter_db_dataset: _IterableBitfountDataset
    ) -> None:
        """Tests that calling the `validation_set` property raises an error.

        The error should be a `DataNotLoadedError` when the data is multi-table because
        it is never loaded.
        """
        with pytest.raises(DataNotLoadedError):
            iter_db_dataset.get_dataset_split(DataSplit.VALIDATION)

    @unit_test
    def test_yield_dataset_split_works_correctly(
        self, dataframe: pd.DataFrame, mock_engine: Mock, mocker: MockerFixture
    ) -> None:
        """Tests that `yield_dataset_split` method works as expected."""
        mock_db_connection = Mock()
        mock_result = Mock()

        class MockPartition:
            """Mock class to represent database result paritions."""

            def __iter__(self) -> Iterator[np.ndarray]:
                """Iterator just returns one set of dataframe values."""
                yield dataframe.values

        mock_result.partitions.return_value = MockPartition()
        mock_result.keys.return_value = dataframe.columns
        mock_result.scalar_one.return_value = DATASET_ROW_COUNT
        mock_engine.execution_options.return_value = mock_engine
        mock_db_connection.execute.return_value = mock_result

        # Creates a multitable DatabaseConnection object
        db_conn = DatabaseConnection(
            mock_engine,
            table_names=["dummy_data", "dummy_data_2"],
        )
        # Mocks `connect` method and resulting context manager on SQLAlchemy Engine
        mocker.patch.object(
            db_conn.con, "connect"
        ).return_value.__enter__.return_value = mock_db_connection

        # Creates DatabaseSource
        ds = DatabaseSource(db_conn, seed=420)
        ds.load_data(table_name="dummy_data")
        mocker.patch.object(_IterableBitfountDataset, "_set_column_name_attributes")
        dataset = _IterableBitfountDataset(ds, Mock(), Mock(), Mock(), Mock())

        # Iterates over datasource split
        df = next(
            dataset.yield_dataset_split(DataSplit.TRAIN)
        )  # First output of iterator
        assert isinstance(df, pd.DataFrame)

        # Makes assertions on call stack
        # Ignoring mypy errors because `connect` has been patched to return a Mock
        db_conn.con.connect.assert_called()  # type: ignore[attr-defined] # Reason: see above # noqa: B950
        db_conn.con.connect.return_value.__enter__.assert_called()  # type: ignore[attr-defined] # Reason: see above # noqa: B950
        db_conn.con.execution_options.assert_called_once()  # type: ignore[attr-defined] # Reason: see above # noqa: B950
        mock_db_connection.execute.assert_called()
        mock_result.partitions.assert_called_once()
        mock_result.keys.assert_called_once()
        mock_result.scalar_one.assert_called_once()

    @integration_test
    def test_database_multi_table_input_table_name(
        self, db_session: sqlalchemy.engine.base.Engine, mocker: MockerFixture
    ) -> None:
        """Checks DatabaseSource initialises correctly with `DatabaseConnection`.

        Multi-table database connection, load single table. Check that table is iterated
        instead of loaded.
        """
        db_conn = DatabaseConnection(
            db_session, table_names=["dummy_data", "dummy_data_2"]
        )

        database_partition_size = 10
        datasource = DatabaseSource(
            db_conn, seed=420, database_partition_size=database_partition_size
        )

        # Test when load_data is called without query
        # DatabaseSource has no data attribute
        datasource.load_data()
        assert not datasource._data_is_loaded
        # Test when load_data is called WITH query
        # DatabaseSource has no data attribute
        table_name = "dummy_data"
        datasource.load_data(table_name=table_name)
        assert not datasource._data_is_loaded  # data is not set

        mocker.patch.object(_IterableBitfountDataset, "_set_column_name_attributes")
        dataset = _IterableBitfountDataset(datasource, Mock(), Mock(), Mock(), Mock())

        iterator = dataset.yield_dataset_split(split=DataSplit.TRAIN)
        df = next(iterator)
        assert len(df) == database_partition_size

    @integration_test
    @pytest.mark.parametrize(
        "query",
        [
            'SELECT "Date", "TARGET" FROM blah',
            'SELECT "invalid", FROM blah',
            '"invalid" from blah',
        ],
    )
    def test_database_query_sql_error(
        self,
        db_session: sqlalchemy.engine.base.Engine,
        mocker: MockerFixture,
        query: str,
    ) -> None:
        """Checks DatabaseSource raises sqlalchemy error."""
        db_conn = DatabaseConnection(
            db_session,
            table_names=["dummy_data", "dummy_data_2"],
        )

        datasource = DatabaseSource(db_conn, seed=420)
        datasource.load_data(sql_query=query)

        mocker.patch.object(_IterableBitfountDataset, "_set_column_name_attributes")
        dataset = _IterableBitfountDataset(datasource, Mock(), Mock(), Mock(), Mock())

        with pytest.raises(sqlalchemy.exc.ProgrammingError):
            next(dataset.yield_dataset_split(split=DataSplit.TRAIN))

    @integration_test
    def test_database_multi_table_input(
        self, db_session: sqlalchemy.engine.base.Engine, mocker: MockerFixture
    ) -> None:
        """Checks DatabaseSource initialises correctly with `DatabaseConnection`.

        Multi-table database connection.
        """
        db_conn = DatabaseConnection(
            db_session, table_names=["dummy_data", "dummy_data_2"]
        )

        datasource = DatabaseSource(
            db_conn, seed=420, data_splitter=PercentageSplitter(0, 0)
        )

        # Test when load_data is called without query
        # DatabaseSource has no data attribute
        datasource.load_data()
        assert datasource.multi_table
        assert not datasource._data_is_loaded
        # Test when load_data is called WITH query
        # DatabaseSource has no data attribute
        query = "SELECT 'Date', 'TARGET' FROM dummy_data"
        datasource.load_data(sql_query=query)
        assert not datasource._data_is_loaded
        assert isinstance(datasource, DatabaseSource)
        expected_output = pd.read_sql(
            f"{query} LIMIT {datasource.database_partition_size}",
            con=db_conn.con,
        )
        mocker.patch.object(_IterableBitfountDataset, "_set_column_name_attributes")
        dataset = _IterableBitfountDataset(datasource, Mock(), Mock(), Mock(), Mock())

        pd.testing.assert_frame_equal(
            next(dataset.yield_dataset_split(split=DataSplit.TRAIN)), expected_output
        )

    @integration_test
    def test_database_single_table_input(
        self, db_session: sqlalchemy.engine.base.Engine, mocker: MockerFixture
    ) -> None:
        """Checks DatabaseSource initialises correctly with `DatabaseConnection`.

        Single table database connection.
        """
        db_conn = DatabaseConnection(db_session, table_names=["dummy_data"])
        datasource = DatabaseSource(db_conn, seed=420)
        datasource.load_data()
        mocker.patch.object(_IterableBitfountDataset, "_set_column_name_attributes")
        dataset = _IterableBitfountDataset(datasource, Mock(), Mock(), Mock(), Mock())
        dataset._split_data()
        assert datasource.data is not None
        assert not datasource.multi_table
        assert datasource._train_idxs is not None
        assert datasource._validation_idxs is not None
        assert datasource._test_idxs is not None
        assert len(datasource._train_idxs) + len(datasource._validation_idxs) + len(
            datasource._test_idxs
        ) == len(datasource.data)

    @unit_test
    def test_mock_database_single_table_input(
        self, mock_engine: Mock, mock_pandas_read_sql_query: None, mocker: MockerFixture
    ) -> None:
        """Checks DatabaseSource initialises correctly with `DatabaseConnection`.

        Mock single table database connection.
        """
        db_conn = DatabaseConnection(mock_engine, table_names=["dummy_data"])
        datasource = DatabaseSource(db_conn, seed=420)
        datasource.load_data()
        mocker.patch.object(_IterableBitfountDataset, "_set_column_name_attributes")
        dataset = _IterableBitfountDataset(datasource, Mock(), Mock(), Mock(), Mock())
        dataset._split_data()
        assert datasource.data is not None
        assert not datasource.multi_table
        assert datasource._train_idxs is not None
        assert datasource._validation_idxs is not None
        assert datasource._test_idxs is not None
        assert len(datasource._train_idxs) + len(datasource._validation_idxs) + len(
            datasource._test_idxs
        ) == len(datasource.data)

    @integration_test
    def test_database_query_input(
        self, db_session: sqlalchemy.engine.base.Engine, mocker: MockerFixture
    ) -> None:
        """Checks DatabaseSource initialises correctly with `DatabaseConnection`.

        Query database connection.
        """
        db_conn = DatabaseConnection(
            db_session,
            query="""
            SELECT *
            FROM dummy_data d1
            LEFT JOIN dummy_data_2 d2
            ON 'd1.Date' = 'd2.Date'
            """,
        )
        datasource = DatabaseSource(db_conn, seed=420)
        datasource.load_data()

        mocker.patch.object(_IterableBitfountDataset, "_set_column_name_attributes")
        dataset = _IterableBitfountDataset(datasource, Mock(), Mock(), Mock(), Mock())
        dataset._split_data()
        assert not datasource.multi_table
        assert datasource.data is not None
        assert datasource._train_idxs is not None
        assert datasource._validation_idxs is not None
        assert datasource._test_idxs is not None
        assert len(datasource._train_idxs) + len(datasource._validation_idxs) + len(
            datasource._test_idxs
        ) == len(datasource.data)



"""Results Only protocol."""
from __future__ import annotations

import os
import time
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    List,
    Mapping,
    Optional,
    Protocol,
    Union,
    cast,
    overload,
    runtime_checkable,
)

from bitfount.data.datasources.base_source import BaseSource
from bitfount.federated.aggregators.base import (
    _AggregatorWorkerFactory,
    _BaseAggregatorFactory,
    _BaseModellerAggregator,
    _BaseWorkerAggregator,
)
from bitfount.federated.aggregators.base import registry as aggregators_registry
from bitfount.federated.aggregators.secure import _InterPodAggregatorWorkerFactory
from bitfount.federated.algorithms.base import registry as algorithms_registry
from bitfount.federated.algorithms.model_algorithms.base import (
    _BaseModellerModelAlgorithm,
    _BaseWorkerModelAlgorithm,
)
from bitfount.federated.helper import _create_aggregator
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.model_reference import BitfountModelReference
from bitfount.federated.pod_vitals import _PodVitals
from bitfount.federated.privacy.differential import DPPodConfig
from bitfount.federated.protocols.base import (
    _BaseCompatibleAlgoFactory,
    _BaseModellerProtocol,
    _BaseProtocolFactory,
    _BaseWorkerProtocol,
)
from bitfount.federated.transport.modeller_transport import (
    _ModellerMailbox,
    _send_model_parameters,
)
from bitfount.federated.transport.worker_transport import (
    _get_model_parameters,
    _InterPodWorkerMailbox,
    _WorkerMailbox,
)
from bitfount.federated.types import SerializedProtocol
from bitfount.schemas.utils import bf_dump
from bitfount.types import (
    T_FIELDS_DICT,
    T_NESTED_FIELDS,
    T_OPTIONAL_NESTED,
    DistributedModelProtocol,
    _SerializedWeights,
    _StrAnyDict,
    _Weights,
)
from bitfount.utils import delegates

if TYPE_CHECKING:
    from bitfount.hub.api import BitfountHub

logger = _get_federated_logger(__name__)


@runtime_checkable
class _ResultsOnlyCompatibleModellerAlgorithm(Protocol):
    """Defines modeller-side algorithm compatibility."""

    def initialise(
        self,
        **kwargs: Any,
    ) -> None:
        """Initialises the modeller-side algorithm."""
        ...

    def run(self, results: Mapping[str, Any]) -> _StrAnyDict:
        """Runs the modeller-side algorithm."""
        ...


@runtime_checkable
class _ResultsOnlyCompatibleWorkerAlgorithm(Protocol):
    """Defines worker-side algorithm compatibility."""

    @overload
    def initialise(
        self,
        datasource: BaseSource,
        **kwargs: Any,
    ) -> None:
        """Initialises the worker-side algorithm without DP."""
        ...

    @overload
    def initialise(
        self,
        datasource: BaseSource,
        pod_dp: Optional[DPPodConfig] = None,
        **kwargs: Any,
    ) -> None:
        """Initialises the worker-side algorithm with DP."""
        ...

    def initialise(
        self,
        datasource: BaseSource,
        pod_dp: Optional[DPPodConfig] = None,
        model_params: Optional[_SerializedWeights] = None,
        **kwargs: Any,
    ) -> None:
        """Initialises the worker-side algorithm with DP and/or model parameters."""
        ...


@runtime_checkable
class _ResultsOnlyDataIncompatibleWorkerAlgorithm(
    _ResultsOnlyCompatibleWorkerAlgorithm, Protocol
):
    """Defines worker-side algorithm compatibility without datasource."""

    def run(self) -> Any:
        """Runs the worker-side algorithm."""
        ...


@runtime_checkable
class _ResultsOnlyDataCompatibleWorkerAlgorithm(
    _ResultsOnlyCompatibleWorkerAlgorithm, Protocol
):
    """Defines worker-side algorithm compatibility with datasource needed."""

    def run(self, data: BaseSource) -> Any:
        """Runs the worker-side algorithm."""
        ...


class _ModellerSide(_BaseModellerProtocol):
    """Modeller side of the ResultsOnly protocol."""

    algorithm: _ResultsOnlyCompatibleModellerAlgorithm
    aggregator: Optional[_BaseModellerAggregator]

    def __init__(
        self,
        *,
        algorithm: _ResultsOnlyCompatibleModellerAlgorithm,
        aggregator: Optional[_BaseModellerAggregator],
        mailbox: _ModellerMailbox,
        **kwargs: Any,
    ):
        super().__init__(algorithm=algorithm, mailbox=mailbox, **kwargs)
        self.aggregator = aggregator

    async def _send_parameters(self, new_parameters: _SerializedWeights) -> None:
        """Sends central model parameters to workers."""
        logger.debug("Sending global parameters to workers")
        await _send_model_parameters(new_parameters, self.mailbox)

    async def run(
        self,
        **kwargs: Any,
    ) -> Union[List[Any], Any]:
        """Runs Modeller side of the protocol."""
        self.algorithm.initialise()
        if isinstance(self.algorithm, _BaseModellerModelAlgorithm):
            initial_parameters: _Weights = self.algorithm.model.get_param_states()
            serialized_params = self.algorithm.model.serialize_params(
                initial_parameters
            )
            await self._send_parameters(serialized_params)
        eval_results = await self.mailbox.get_evaluation_results_from_workers()
        logger.info("Results received from Pods.")
        await self.mailbox.send_task_complete_message()

        modeller_results = self.algorithm.run(eval_results)

        if self.aggregator:
            return self.aggregator.run(modeller_results)

        return modeller_results


class _WorkerSide(_BaseWorkerProtocol):
    """Worker side of the ResultsOnly protocol."""

    algorithm: Union[
        _ResultsOnlyDataIncompatibleWorkerAlgorithm,
        _ResultsOnlyDataCompatibleWorkerAlgorithm,
    ]
    aggregator: Optional[_BaseWorkerAggregator]

    def __init__(
        self,
        *,
        algorithm: Union[
            _ResultsOnlyDataIncompatibleWorkerAlgorithm,
            _ResultsOnlyDataCompatibleWorkerAlgorithm,
        ],
        aggregator: Optional[_BaseWorkerAggregator],
        mailbox: _WorkerMailbox,
        **kwargs: Any,
    ):
        super().__init__(algorithm=algorithm, mailbox=mailbox, **kwargs)
        self.aggregator = aggregator

    async def _receive_parameters(self) -> _SerializedWeights:
        """Receives new global model parameters."""
        logger.debug("Receiving global parameters")
        return await _get_model_parameters(self.mailbox)

    async def run(
        self,
        datasource: BaseSource,
        pod_dp: Optional[DPPodConfig] = None,
        pod_vitals: Optional[_PodVitals] = None,
        pod_identifier: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Runs Worker side of the protocol."""
        if isinstance(self.algorithm, _BaseWorkerModelAlgorithm):
            # Receive the initial parameters here from the Modeller
            model_params = await self._receive_parameters()
        else:
            model_params = None

        # Pass model_params to initialise where we update the parameters
        self.algorithm.initialise(
            datasource=datasource,
            pod_dp=pod_dp,
            pod_identifier=pod_identifier,
            model_params=model_params,
        )

        if pod_vitals:
            pod_vitals.last_task_execution_time = time.time()
        try:
            self.algorithm = cast(
                _ResultsOnlyDataCompatibleWorkerAlgorithm, self.algorithm
            )
            results = self.algorithm.run(data=datasource)
        except TypeError:
            self.algorithm = cast(
                _ResultsOnlyDataIncompatibleWorkerAlgorithm, self.algorithm
            )
            results = self.algorithm.run()

        if self.aggregator:
            logger.debug("Aggregating results...")
            results = await self.aggregator.run(results)
            logger.debug("Aggregation complete.")

        await self.mailbox.send_evaluation_results(results)
        await self.mailbox.get_task_complete_update()


@runtime_checkable
class _ResultsOnlyCompatibleAlgoFactory(_BaseCompatibleAlgoFactory, Protocol):
    """Defines algo factory compatibility."""

    def modeller(self, **kwargs: Any) -> _ResultsOnlyCompatibleModellerAlgorithm:
        """Create a modeller-side algorithm."""
        ...


@runtime_checkable
class _ResultsOnlyCompatibleAlgoFactory_(_ResultsOnlyCompatibleAlgoFactory, Protocol):
    """Defines algo factory compatibility."""

    def worker(
        self, **kwargs: Any
    ) -> Union[
        _ResultsOnlyDataIncompatibleWorkerAlgorithm,
        _ResultsOnlyDataCompatibleWorkerAlgorithm,
    ]:
        """Create a worker-side algorithm."""
        ...


@runtime_checkable
class _ResultsOnlyCompatibleModelAlgoFactory(
    _ResultsOnlyCompatibleAlgoFactory, Protocol
):
    """Defines algo factory compatibility."""

    model: Union[DistributedModelProtocol, BitfountModelReference]
    pretrained_file: Optional[Union[str, os.PathLike]] = None

    def worker(
        self, hub: BitfountHub, **kwargs: Any
    ) -> Union[
        _ResultsOnlyDataIncompatibleWorkerAlgorithm,
        _ResultsOnlyDataCompatibleWorkerAlgorithm,
    ]:
        """Create a worker-side algorithm."""
        ...


@delegates()
class ResultsOnly(_BaseProtocolFactory):
    """Simply returns the results from the provided algorithm.

    This protocol is the most permissive protocol and only involves one round of
    communication. It simply runs the algorithm on the `Pod`(s) and returns the
    results as a list (one element for every pod) unless an aggregator is specified.

    Args:
        algorithm: The algorithm to run.
        aggregator: The aggregator to use for updating the algorithm results across all
            Pods participating in the task.  This argument takes priority over the
            `secure_aggregation` argument.
        secure_aggregation: Whether to use secure aggregation. This argument is
            overridden by the `aggregator` argument.

    Attributes:
        name: The name of the protocol.
        algorithm: The algorithm to run. This must be compatible with the `ResultsOnly`
            protocol.
        aggregator: The aggregator to use for updating the algorithm results.

    Raises:
        TypeError: If the `algorithm` is not compatible with the protocol.
    """

    # TODO: [BIT-1047] Consider separating this protocol into two separate protocols
    # for each algorithm. The algorithms may not be similar enough to benefit
    # from sharing one protocol.

    algorithm: Union[
        _ResultsOnlyCompatibleAlgoFactory_, _ResultsOnlyCompatibleModelAlgoFactory
    ]
    fields_dict: ClassVar[T_FIELDS_DICT] = {}
    nested_fields: ClassVar[T_NESTED_FIELDS] = {
        "algorithm": algorithms_registry,
        "aggregator": aggregators_registry,
    }
    optional_nested: ClassVar[T_OPTIONAL_NESTED] = ["aggregator"]

    def __init__(
        self,
        *,
        algorithm: Union[
            _ResultsOnlyCompatibleAlgoFactory_, _ResultsOnlyCompatibleModelAlgoFactory
        ],
        aggregator: Optional[_BaseAggregatorFactory] = None,
        secure_aggregation: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(algorithm=algorithm, **kwargs)
        self.aggregator: Optional[_BaseAggregatorFactory] = None

        if aggregator:
            self.aggregator = aggregator
        elif secure_aggregation:
            self.aggregator = _create_aggregator(secure_aggregation=secure_aggregation)
        else:
            logger.info("No aggregator specified. Will return a dictionary of results.")

    @classmethod
    def _validate_algorithm(
        cls,
        algorithm: _BaseCompatibleAlgoFactory,
    ) -> None:
        """Checks that `algorithm` is compatible with the protocol."""
        if not isinstance(
            algorithm,
            (
                _ResultsOnlyCompatibleAlgoFactory_,
                _ResultsOnlyCompatibleModelAlgoFactory,
            ),
        ):
            raise TypeError(
                f"The {cls.__name__} protocol does not support "
                + f"the {type(algorithm).__name__} algorithm.",
            )

    def dump(self) -> SerializedProtocol:
        """Returns the JSON-serializable representation of the protocol."""
        return cast(SerializedProtocol, bf_dump(self))

    def modeller(self, mailbox: _ModellerMailbox, **kwargs: Any) -> _ModellerSide:
        """Returns the modeller side of the ResultsOnly protocol."""
        if isinstance(self.algorithm, _ResultsOnlyCompatibleModelAlgoFactory):
            algorithm = self.algorithm.modeller(
                pretrained_file=self.algorithm.pretrained_file
            )
        else:
            algorithm = self.algorithm.modeller()
        return _ModellerSide(
            algorithm=algorithm,
            aggregator=self.aggregator.modeller() if self.aggregator else None,
            mailbox=mailbox,
            **kwargs,
        )

    def worker(
        self, mailbox: _WorkerMailbox, hub: BitfountHub, **kwargs: Any
    ) -> _WorkerSide:
        """Returns the worker side of the ResultsOnly protocol.

        Raises:
            TypeError: If the mailbox is not compatible with the aggregator.
        """
        worker_agg: Optional[_BaseWorkerAggregator] = None
        if self.aggregator is not None:
            if isinstance(self.aggregator, _AggregatorWorkerFactory):
                worker_agg = self.aggregator.worker()
            elif isinstance(self.aggregator, _InterPodAggregatorWorkerFactory):
                if not isinstance(mailbox, _InterPodWorkerMailbox):
                    raise TypeError(
                        "Inter-pod aggregators require an inter-pod worker mailbox."
                    )
                worker_agg = self.aggregator.worker(mailbox=mailbox)
            else:
                raise TypeError(
                    f"Unrecognised aggregator factory ({type(self.aggregator)}); "
                    f"unable to determine how to call .worker() factory method."
                )

        return _WorkerSide(
            algorithm=self.algorithm.worker(hub=hub),
            aggregator=worker_agg,
            mailbox=mailbox,
            **kwargs,
        )



"""Tests for base model classes."""
from __future__ import annotations

from typing import Any, cast
from unittest.mock import Mock

from marshmallow import fields
from pytest import fixture

from bitfount.data.datastructure import DataStructure
from bitfount.models.base_models import (
    MAIN_MODEL_REGISTRY,
    EarlyStopping,
    FeedForwardModelStructure,
    NeuralNetworkMixIn,
    NeuralNetworkModelStructure,
    Optimizer,
    Scheduler,
)
from bitfount.schemas.utils import bf_dump, bf_load
from tests.utils.helper import create_datastructure, unit_test


class _NeuralNetworkMixIn_(NeuralNetworkMixIn):
    def __init__(self, datastructure: DataStructure, **kwargs: Any):
        self.datastructure = datastructure
        self.class_name = type(self).__name__

        super().__init__(**kwargs)
        fields_dict = {  # noqa: F841
            "datastructure": fields.Raw(),
        }

    @staticmethod
    def _get_optimizer(  # type:ignore[override] # reason, used just for testing
        optimizer: Optimizer,
    ) -> None:
        return None

    @staticmethod
    def _get_scheduler(  # type:ignore[override] # reason, used just for testing
        scheduler: Scheduler,
    ) -> None:
        return None


@unit_test
class TestNeuralNetworkMixIn:
    """Tests for the NeuralNetworkMixIn."""

    @fixture
    def model_structure(self) -> NeuralNetworkModelStructure:
        """Model structure fixture."""
        return NeuralNetworkModelStructure()

    @fixture
    def mock_datastructure(self) -> Mock:
        """Mock datastructure fixture."""
        return Mock(multihead_col=None)

    def test_scheduler_structure_works_with_params(
        self, mock_datastructure: Mock, model_structure: NeuralNetworkModelStructure
    ) -> None:
        """Tests the scheduler loads parameters correctly."""
        nn = _NeuralNetworkMixIn_(
            datastructure=mock_datastructure,
            model_structure=model_structure,
            epochs=1,
            scheduler=Scheduler("OneCycleLR", {"max_lr": 0.01}),
        )

        model_scheduler = cast(Scheduler, nn.scheduler)
        assert model_scheduler.name == "OneCycleLR"
        assert model_scheduler.params == {"max_lr": 0.01}

    def test_scheduler_structure_works_no_params(
        self, mock_datastructure: Mock, model_structure: NeuralNetworkModelStructure
    ) -> None:
        """Tests the scheduler loads correctly when no scheduler_params given."""
        nn = _NeuralNetworkMixIn_(
            datastructure=mock_datastructure,
            model_structure=model_structure,
            epochs=1,
            scheduler=Scheduler("OneCycleLR"),
        )

        model_scheduler = cast(Scheduler, nn.scheduler)
        assert model_scheduler.name == "OneCycleLR"
        assert model_scheduler.params == {}

    def test_early_stopping(
        self, mock_datastructure: Mock, model_structure: NeuralNetworkModelStructure
    ) -> None:
        """Tests early stopping loads parameters correctly."""
        early_stopping_params = dict(
            monitor="validation_loss",
            min_delta=0.00,
            patience=2,
            verbose=True,
            mode="min",
        )
        nn = _NeuralNetworkMixIn_(
            datastructure=mock_datastructure,
            model_structure=model_structure,
            epochs=10,
            early_stopping=EarlyStopping(early_stopping_params),
        )
        model_es = cast(EarlyStopping, nn.early_stopping)
        assert model_es.params == early_stopping_params
        assert model_es is not None

    def test_optimizer(
        self, mock_datastructure: Mock, model_structure: NeuralNetworkModelStructure
    ) -> None:
        """Tests the optimizer loads parameters correctly."""
        optimizer = Optimizer(name="SGD", params={"lr": 0.1})
        nn = _NeuralNetworkMixIn_(
            datastructure=mock_datastructure,
            model_structure=model_structure,
            epochs=1,
            optimizer=optimizer,
        )
        model_optimizer = nn.optimizer
        assert model_optimizer.name == "SGD"
        assert model_optimizer.params == {"lr": 0.1}

    def test_no_optimizer_no_params(
        self, mock_datastructure: Mock, model_structure: NeuralNetworkModelStructure
    ) -> None:
        """Tests the default optimizer with defined params loads correctly."""
        nn = _NeuralNetworkMixIn_(
            datastructure=mock_datastructure,
            model_structure=model_structure,
            epochs=1,
        )
        model_optimizer = nn.optimizer
        assert model_optimizer.name == "AdamW"
        assert model_optimizer.params == {"lr": 0.01}

    def test_optimizer_no_params(
        self, mock_datastructure: Mock, model_structure: NeuralNetworkModelStructure
    ) -> None:
        """Tests the optimizer loads params correctly with no optimizer_params."""
        optimizer = Optimizer(name="SGD")
        nn = _NeuralNetworkMixIn_(
            datastructure=mock_datastructure,
            model_structure=model_structure,
            epochs=1,
            optimizer=optimizer,
        )
        model_optimizer = nn.optimizer
        assert model_optimizer.name == "SGD"
        assert model_optimizer.params == {}

    def test_serialization_deserialization(self, mock_datastructure: Mock) -> None:
        """Tests (de)serialization of the NeuralNetworkMixin class."""
        nn = _NeuralNetworkMixIn_(
            datastructure=create_datastructure(),
            model_structure=FeedForwardModelStructure(),
            batch_size=1,
            epochs=1,
            steps=None,  # as only one of epochs/steps can be specified
            optimizer=Optimizer(name="blah", params={"blah": 1}),
            scheduler=Scheduler("blah", {"blah": 1}),
            custom_loss_func=None,
            early_stopping=EarlyStopping({"blah": 1}),
        )

        # Dump and reload the class variables
        dumped = bf_dump(nn)
        loaded = bf_load(dumped, MAIN_MODEL_REGISTRY)

        # We compare across __dict__ as NeuralNetworkMixIn does not have
        # an __eq__ method.
        assert loaded.__dict__ == nn.__dict__



"""Main script to run others as subcommands."""
import fire

from .generate_schema import gen_schema
from .run_pod import run as pod_run
from .run_testing import evaluate_model


def main() -> None:
    """Main script entry point."""
    fire.Fire(
        {
            "generate_schema": gen_schema,
            "run_pod": pod_run,
            "run_testing": evaluate_model,
        }
    )


if __name__ == "__main__":
    main()



"""Module containing DataFrameSource class.

DataFrameSource class handles loading data stored in memory in a pandas dataframe.
"""
from __future__ import annotations

import logging
from typing import Any, Dict, Iterable, List, Union

import numpy as np
import pandas as pd

from bitfount.data.datasources.base_source import BaseSource
from bitfount.types import _Dtypes
from bitfount.utils import delegates

logger = logging.getLogger(__name__)


@delegates()
class DataFrameSource(BaseSource):
    """Data source for loading dataframes.

    Args:
        data: The dataframe to be loaded.
    """

    def __init__(self, data: pd.DataFrame, **kwargs: Any):
        super().__init__(**kwargs)
        if isinstance(data, pd.DataFrame):
            self.dataframe: pd.DataFrame = data
        else:
            raise TypeError(
                "Invalid data attribute. "
                "Expected pandas dataframe "
                f"but received :{type(data)}"
            )

    def get_values(
        self, col_names: List[str], **kwargs: Any
    ) -> Dict[str, Iterable[Any]]:
        """Get distinct values from columns in DataFrame dataset.

        Args:
            col_names: The list of the columns whose distinct values should be
                returned.

        Returns:
            The distinct values of the requested column as a mapping from col name to
            a series of distinct values.
        """
        return {col: self.dataframe[col].unique() for col in col_names}

    def get_column(self, col_name: str, **kwargs: Any) -> Union[np.ndarray, pd.Series]:
        """Loads and returns single column from dataframe dataset.

        Args:
            col_name: The name of the column which should be loaded.

        Returns:
            The column request as a series.
        """
        return self.dataframe[col_name]

    def get_data(self, **kwargs: Any) -> pd.DataFrame:
        """Loads and returns datafrom DataFrame dataset.

        Returns:
            A DataFrame-type object which contains the data.
        """
        return self.dataframe

    def get_dtypes(self, **kwargs: Any) -> _Dtypes:
        """Loads and returns the columns and column types from the Dataframe dataset.

        Returns:
            A mapping from column names to column types.
        """
        return self._get_data_dtypes(self.dataframe)

    def __len__(self) -> int:
        return len(self.dataframe)

    @property
    def multi_table(self) -> bool:
        """Attribute to specify whether the datasource is multi table."""
        return False



"""Testcases for all classes in bitfount/models.py."""
import os
from pathlib import Path
from typing import cast
from unittest.mock import Mock

import numpy as np
import pandas as pd
import pytest
from pytest import fixture
from pytest_mock import MockerFixture
from sklearn.linear_model import LinearRegression as sklearnLinearRegression
from sklearn.neighbors import KNeighborsClassifier

from bitfount.data.dataloaders import BitfountDataLoader
from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datastructure import DataStructure
from bitfount.data.schema import BitfountSchema
from bitfount.metrics import MetricCollection
from bitfount.models.base_models import (
    MAIN_MODEL_REGISTRY,
    CNNModelStructure,
    NeuralNetworkModelStructure,
    _model_structure_registry,
)
from bitfount.models.models import LogisticRegressionClassifier, RegBoostRegressor
from bitfount.schemas.utils import bf_dump, bf_load
from bitfount.types import _StrAnyDict
from tests.utils.helper import (
    TABLE_NAME,
    assert_results,
    create_dataset,
    create_datasource,
    create_datastructure,
    create_schema,
    integration_test,
    unit_test,
)

SERIALIZED_MODEL_NAME = "test_model.pickle"


def assert_vars_equal(vars_original: _StrAnyDict, vars_copy: _StrAnyDict) -> None:
    """Asserts both vars() are equal; handles FunctionalForm as well."""
    for variable, value in vars_original.items():
        if isinstance(value, DataStructure):
            assert value.target == vars_copy[variable].target
        elif isinstance(value, BitfountSchema):
            assert isinstance(vars_copy[variable], BitfountSchema)
        else:
            assert value == vars_copy[variable]


@pytest.fixture
def datastructure() -> DataStructure:
    """Fixture for datastructure."""
    return create_datastructure()


@pytest.fixture
def datasource() -> BaseSource:
    """Fixture for datasource."""
    return create_datasource(classification=True)


@pytest.fixture
def schema() -> BitfountSchema:
    """Fixture for schema."""
    return create_schema(classification=True)


@unit_test
class TestNeuralNetworkModelStructure:
    """Test NeuralNetworkModelStructure classes."""

    def test_nn_model_structure_dropout_layer_mismatch(self) -> None:
        """Test ValueError is raised if dropout layers don't match linear layers."""
        with pytest.raises(ValueError):
            NeuralNetworkModelStructure(layers=[100, 50], dropout_probs=[0.1, 0.3, 0.5])

    def test_cnn_model_structure_invalid_pooling_function(self) -> None:
        """Test ValueError is raised if an invalid pooling function is used."""
        with pytest.raises(ValueError):
            CNNModelStructure(pooling_function="notarealpoolingfunction")

    def test_cnn_model_structure_dropout_layer_mismatch(self) -> None:
        """Test ValueError is raised if dropout layers don't match conv layers."""
        with pytest.raises(ValueError):
            CNNModelStructure(layers=[100, 50], dropout_probs=[0.1, 0.3, 0.5])

    def test_cnn_model_structure_serialization(self) -> None:
        """Test CNN Model is serialized correctly."""
        model_structure = CNNModelStructure(pooling_function="avg", padding=5, stride=1)
        dumped_model_structure = bf_dump(model_structure)
        reloaded = bf_load(dumped_model_structure, _model_structure_registry)
        assert reloaded.pooling_function == model_structure.pooling_function
        assert reloaded.padding == model_structure.padding
        assert reloaded.stride == model_structure.stride


class TestLogisticRegression:
    """Test LogisticRegression class from bitfount.models."""

    @integration_test
    def test_fit_and_results(
        self, datasource: BaseSource, datastructure: DataStructure
    ) -> None:
        """Test LogisticRegression fit() and get_results() methods."""
        log_reg = LogisticRegressionClassifier(
            datastructure=datastructure,
            schema=create_schema(classification=True),
            verbose=0,
        )
        log_reg.fit(datasource)
        assert_results(model=log_reg)

    @unit_test
    def test_missing_image_col_raises_error(self) -> None:
        """Tests that a value error when images are not in the datasource."""  # noqa: B950
        datasource = create_datasource(classification=True, image=True)
        datastructure = DataStructure(
            target=["TARGET"], image_cols=["img"], table=TABLE_NAME
        )
        model = LogisticRegressionClassifier(
            datastructure=datastructure,
            schema=create_schema(classification=True),
            verbose=0,
        )
        with pytest.raises(ValueError):
            model.fit(datasource)

    @unit_test
    def test_evaluate_no_test_dl_error(self, datastructure: DataStructure) -> None:
        """Tests that evaluate raises error with no test_dl."""
        model = LogisticRegressionClassifier(
            datastructure=datastructure,
            schema=create_schema(classification=True),
            verbose=-1,
        )
        with pytest.raises(ValueError):
            model.evaluate()

    @unit_test
    def test_serialization(
        self, datasource: BaseSource, datastructure: DataStructure, tmp_path: Path
    ) -> None:
        """Test serialize() and deserialize() methods."""
        log_reg = LogisticRegressionClassifier(
            datastructure=datastructure,
            schema=create_schema(classification=True),
            verbose=0,
        )
        log_reg.fit(datasource)
        log_reg.serialize(tmp_path / SERIALIZED_MODEL_NAME)
        assert os.path.exists(tmp_path / SERIALIZED_MODEL_NAME)

        log_reg = LogisticRegressionClassifier(
            datastructure=datastructure, schema=create_schema(classification=True)
        )
        log_reg.fit(datasource)
        log_reg.deserialize(tmp_path / SERIALIZED_MODEL_NAME)

        assert log_reg.test_dl is not None
        log_reg.evaluate(log_reg.test_dl)


class TestRegBoostRegressor:
    """Test RegBoost model."""

    @fixture
    def mock_ols(self, mocker: MockerFixture) -> Mock:
        """Mock OLS import."""
        mock_ols: Mock = mocker.patch("statsmodels.api.OLS", autospec=True)
        return mock_ols

    @unit_test
    def test_model_tree_structure(self) -> None:
        """Test basic Model tree structure."""
        data = create_dataset(classification=True)
        modeltree = RegBoostRegressor._ModelTree(data.index.to_numpy(), 0)
        assert modeltree.negatives is None
        assert modeltree.positives is None
        assert isinstance(modeltree.model, sklearnLinearRegression)
        assert modeltree.depth == 0
        assert len(modeltree.data_indices) == len(data)

    @unit_test
    def test_model_tree_structure_add_children(self) -> None:
        """Test Model tree structure adding children."""
        data = create_dataset(classification=True)
        modeltree = RegBoostRegressor._ModelTree(data.index.to_numpy(), 0)
        child = modeltree.add_child(
            RegBoostRegressor._ModelTreeSide.NEGATIVE, np.array([1, 2, 3, 4, 5])
        )
        assert modeltree.negatives is not None
        assert modeltree.negatives.depth == 1
        assert child == modeltree.negatives
        assert isinstance(modeltree.negatives.model, sklearnLinearRegression)
        assert len(modeltree.negatives.data_indices) == 5
        assert modeltree.positives is None

        child2 = modeltree.add_child(
            RegBoostRegressor._ModelTreeSide.POSITIVE, np.array([1, 2, 3, 4])
        )

        assert modeltree.positives is not None
        assert modeltree.positives.depth == 1
        assert child2 == modeltree.positives
        assert child != child2
        assert isinstance(modeltree.positives.model, sklearnLinearRegression)
        assert len(modeltree.positives.data_indices) == 4
        assert modeltree.negatives is not None

    @unit_test
    @pytest.mark.parametrize("max_depth", [1, 5, 10])
    def test_model_tree_building(
        self, datasource: BaseSource, datastructure: DataStructure, max_depth: int
    ) -> None:
        """Test building of Model tree."""
        reg = RegBoostRegressor(
            datastructure=datastructure, schema=BitfountSchema(), max_depth=max_depth
        )
        reg.fit(datasource)

        assert isinstance(reg._model, RegBoostRegressor._ModelTree)

        def find_tree_max_depth(model: RegBoostRegressor._ModelTree) -> int:
            depths = []
            for model_ in [model.positives, model.negatives]:
                if model_ is not None:
                    # Assert data indices are a subset of parent and >= minimum required
                    assert len(model_.data_indices) >= reg.min_data_points_per_node
                    assert set(model_.data_indices).issubset(set(model.data_indices))
                    # Assert features are a subset of parent
                    assert len(model_.features) <= len(model.features)
                    assert set(model_.features).issubset(set(model.features))
                    depths.append(find_tree_max_depth(model_))
                else:
                    depths.append(model.depth)
            return max(depths)

        depth = find_tree_max_depth(reg._model)
        assert depth <= reg.max_depth

    @unit_test
    def test_fit_and_evaluate_model_tree(
        self,
        datasource: BaseSource,
        datastructure: DataStructure,
        mocker: MockerFixture,
        schema: BitfountSchema,
    ) -> None:
        """Test fitting and evaluating of model tree with very basic dummy data.

        Dummy data consists of 6 data points which can be separated into two exactly
        linear relationships of 3 data points each such that our predictions should be
        precisely accurate after one splitting of the decision tree.
        """
        model = RegBoostRegressor(
            datastructure=datastructure,
            schema=schema,
            max_depth=1,
            learning_rate=0.1,
            min_data_points_per_node=2,
        )

        # Patch out train dataloader
        model.train_dl = Mock()
        X = pd.DataFrame({"x": [3, 5, 7, 2, 4, 6]})
        y = pd.DataFrame({"TARGET": [2, 3, 4, 5, 6, 7]})
        positive_indices = [3, 4, 5]
        negative_indices = [0, 1, 2]
        model.train_dl.get_x_dataframe = Mock(return_value=X)
        model.train_dl.get_y_dataframe = Mock(return_value=y)

        mock_fit = mocker.patch.object(
            RegBoostRegressor._ModelTree,
            "fit",
            autospec=True,
            side_effect=lambda self_, X, y: self_.model.fit(X, y),
        )
        mock_predict = mocker.patch.object(
            RegBoostRegressor._ModelTree,
            "predict",
            autospec=True,
            side_effect=lambda self_, X: self_.model.predict(X),
        )

        modeltree = RegBoostRegressor._ModelTree(np.array(range(6)), 0)
        model._fit_model_tree(modeltree)

        assert modeltree.positives is not None
        assert modeltree.negatives is not None
        assert modeltree.positives.data_indices.tolist() == positive_indices
        assert modeltree.negatives.data_indices.tolist() == negative_indices
        assert modeltree.negatives.is_leaf
        assert modeltree.positives.is_leaf
        mock_predict.assert_called_once()
        pd.testing.assert_frame_equal(mock_predict.call_args[0][1], X)

        # One for each of root node, positive child and negative child
        assert mock_fit.call_count == 3

        # ROOT NODE
        # First call, second positional arg (positional args include 'self')
        pd.testing.assert_frame_equal(mock_fit.call_args_list[0][0][1], X)
        # First call, third positional arg
        np.testing.assert_array_equal(
            mock_fit.call_args_list[0][0][2], y.TARGET.to_numpy()
        )
        # POSITIVE CHILD
        # Second call, second positional arg
        pd.testing.assert_frame_equal(
            mock_fit.call_args_list[1][0][1], X.loc[positive_indices]
        )
        # Second call, third positional arg
        np.testing.assert_array_equal(
            mock_fit.call_args_list[1][0][2],
            # new y values (target - (learning rate * parent predictions))
            np.array([4.6, 5.56, 6.52], dtype=np.float32),
        )
        # NEGATIVE CHILD
        # Third call, second positional arg
        pd.testing.assert_frame_equal(
            mock_fit.call_args_list[2][0][1], X.loc[negative_indices]
        )
        # Third call, third positional arg
        np.testing.assert_array_equal(
            mock_fit.call_args_list[2][0][2],
            # new y values (target - (learning rate * parent predictions))
            np.array([1.58, 2.54, 3.5], dtype=np.float32),
        )

        preds = model._evaluate_model_tree(modeltree, X, 1)
        predictions: np.ndarray = model._aggregate_model_predictions(preds)
        # Assert that our model predictions are precisely accurate
        np.testing.assert_almost_equal(
            predictions, y.TARGET.to_numpy(dtype=np.float32), decimal=4
        )

    @unit_test
    def test_serialization(
        self, datasource: BaseSource, datastructure: DataStructure, tmp_path: Path
    ) -> None:
        """Test serialize() and deserialize() methods."""
        model = RegBoostRegressor(datastructure=datastructure, schema=BitfountSchema())
        model.fit(datasource)
        model.serialize(tmp_path / SERIALIZED_MODEL_NAME)
        assert os.path.exists(tmp_path / SERIALIZED_MODEL_NAME)
        model = RegBoostRegressor(datastructure=datastructure, schema=BitfountSchema())
        model.fit(datasource)
        model.deserialize(tmp_path / SERIALIZED_MODEL_NAME)
        model.evaluate(k=5)

    @unit_test
    def test_append_model_predictions(self) -> None:
        """Test appending of model predictions is in correct order."""
        parent_preds = [
            [0.1, 0.2, 0.3],
            [0.4, 0.5, 0.6],
        ]
        child_preds = [
            [0.7],
            [0.8],
        ]
        indices = [True, False]
        parent_preds = RegBoostRegressor._append_model_predictions(
            parent_preds, child_preds, indices
        )
        assert parent_preds == [
            [0.1, 0.2, 0.3, 0.7],
            [0.4, 0.5, 0.6],
        ]

    @unit_test
    def test_aggregate_model_predictions(
        self, datasource: BaseSource, datastructure: DataStructure
    ) -> None:
        """Test models predictions are aggregated correctly."""
        model = RegBoostRegressor(
            datastructure=datastructure, schema=BitfountSchema(), learning_rate=0.1
        )
        preds = [
            [0.1, 0.2, 0.3],
            [0.4, 0.5, 0.6],
        ]
        agg_preds = model._aggregate_model_predictions(preds)
        # All bar last element in array are multiplied by learning rate and then summed
        np.testing.assert_array_equal(agg_preds, np.array([0.33, 0.69]))

    @unit_test
    def test_model_classifier(
        self, datasource: BaseSource, datastructure: DataStructure
    ) -> None:
        """Test fitting and evaluating of model classifier."""
        model = RegBoostRegressor(datastructure=datastructure, schema=BitfountSchema())
        model.fit(datasource)
        assert isinstance(model._model, RegBoostRegressor._ModelTree)
        assert model._model.classifier is None
        model._fit_model_tree_classifier(model._model, k=5)
        assert isinstance(model._model.classifier, KNeighborsClassifier)
        test_df = cast(BitfountDataLoader, model.validation_dl).get_x_dataframe()
        assert not isinstance(test_df, tuple)
        preds = model._eval_model_tree_classifier(
            model._model, test_df[model._model.features]
        )
        assert len(preds) == len(test_df)

        # Set negatives to None and assert we get positive classes only
        temp_model = model._model.negatives
        model._model.negatives = None
        model._model.classifier = None
        preds = model._eval_model_tree_classifier(
            model._model, test_df[model._model.features]
        )
        np.testing.assert_array_equal(preds, np.ones(len(test_df), dtype=int))

        # Set positives to None and assert we get negative classes only
        model._model.negatives = temp_model
        model._model.positives = None
        model._model.classifier = None
        preds = model._eval_model_tree_classifier(
            model._model, test_df[model._model.features]
        )
        np.testing.assert_array_equal(preds, np.zeros(len(test_df), dtype=int))

    @unit_test
    def test_missing_image_col_raises_error(self) -> None:
        """Tests that a value error when images are not in the datasource."""  # noqa: B950
        datasource = create_datasource(classification=True, image=True)
        datastructure = DataStructure(
            target=["TARGET"], image_cols=["img"], table=TABLE_NAME
        )
        model = RegBoostRegressor(
            datastructure=datastructure, schema=BitfountSchema(), verbose=0
        )
        with pytest.raises(ValueError):
            model.fit(datasource)

    @unit_test
    def test_perform_stepwise_regression(
        self,
        datasource: BaseSource,
        datastructure: DataStructure,
        mocker: MockerFixture,
    ) -> None:
        """Test stepwise regression function."""
        model = RegBoostRegressor(
            datastructure=datastructure,
            schema=BitfountSchema(),
            stepwise_regression="forward",
        )

        fwd = mocker.patch.object(
            RegBoostRegressor,
            "_forward_stepwise_regression",
            autospec=True,
            side_effect=lambda x, y, z: ["test"],
        )
        bwd = mocker.patch.object(
            RegBoostRegressor,
            "_backward_stepwise_regression",
            autospec=True,
            side_effect=lambda x, y, z: ["test"],
        )

        X = pd.DataFrame({"x": [3, 5, 7, 2, 4, 6]})
        y = np.array([2, 3, 4, 5, 6, 7])

        # Test forward stepwise regression
        features = model._perform_stepwise_regression(X, y)
        fwd.assert_called_once()
        bwd.assert_not_called()
        assert features == ["test"]

        # Test backward stepwise regression
        model.stepwise_regression = "backward"
        fwd.reset_mock()
        features = model._perform_stepwise_regression(X, y)
        bwd.assert_called_once()
        fwd.assert_not_called()
        assert features == ["test"]

        # Test stepwise regression not recognised
        with pytest.raises(ValueError):
            model.stepwise_regression = "blah"  # type: ignore[assignment] # Reason: purpose of test # noqa: B950
            model._perform_stepwise_regression(X, y)

    @unit_test
    def test_forward_stepwise_regression(self, mock_ols: Mock) -> None:
        """Test forward stepwise regression."""
        ols_model = Mock()
        mock_ols.return_value = ols_model
        # Cycles through features, adds one and then cycles through remaining features
        ols_model.fit.side_effect = [
            Mock(pvalues={"b": 0.02}),
            Mock(pvalues={"a": 0.11}),
            Mock(pvalues={"b": 0.09, "a": 0.11}),
        ]

        X = pd.DataFrame({"b": [3, 5, 7, 2, 4, 6], "a": [3, 5, 7, 2, 4, 6]})
        y = np.array([2, 3, 4, 5, 6, 7])
        features = RegBoostRegressor._forward_stepwise_regression(X, y, 0.1)
        assert features == ["b"]
        assert mock_ols.call_count == 3

    @unit_test
    def test_backward_stepwise_regression(self, mock_ols: Mock) -> None:
        """Test backward stepwise regression."""
        ols_model = Mock()
        mock_ols.return_value = ols_model
        # Cycles through features, adds one and then cycles through remaining features
        ols_model.fit.side_effect = [
            Mock(pvalues=pd.Series([0, 0.09, 0.11], index=["ignore-me", "b", "a"])),
        ]

        X = pd.DataFrame({"b": [3, 5, 7, 2, 4, 6], "a": [3, 5, 7, 2, 4, 6]})
        y = np.array([2, 3, 4, 5, 6, 7])
        features = RegBoostRegressor._backward_stepwise_regression(X, y, 0.1)
        assert features == ["b"]
        assert mock_ols.call_count == 1

    @pytest.mark.parametrize("max_depth", [1, 3])
    @unit_test
    def test_model_visualisation(
        self, datasource: BaseSource, datastructure: DataStructure, max_depth: int
    ) -> None:
        """Test model visualisation works."""
        model = RegBoostRegressor(
            datastructure=datastructure,
            schema=BitfountSchema(),
            max_depth=max_depth,
        )
        model.fit(datasource)
        string_display = str(model._model)
        num_nodes = sum([2**i for i in range(1, max_depth + 1)]) + 1
        num_tabs = sum([i * (2**i) for i in range(1, max_depth + 1)])
        assert string_display.count("\n") == num_nodes
        assert string_display.count("\t") == num_tabs
        assert string_display.count("Depth") == num_nodes
        assert string_display.count("Features") == num_nodes
        assert string_display.count("Data") == num_nodes

    @unit_test
    def test_regboost_evaluate_no_test_dl_error(
        self, datastructure: DataStructure
    ) -> None:
        """Tests that evaluate raises error with no test_dl."""
        model = RegBoostRegressor(
            datastructure=datastructure,
            schema=BitfountSchema(),
            min_data_points_per_node=5,
            learning_rate=0.5,
            stepwise_regression_threshold=0.01,
        )
        with pytest.raises(ValueError):
            model.evaluate()

    @integration_test
    def test_fit_and_results(
        self, datasource: BaseSource, datastructure: DataStructure
    ) -> None:
        """Test fit() and evaluate() methods as classifier outputs."""
        model = RegBoostRegressor(
            datastructure=datastructure,
            schema=BitfountSchema(),
            min_data_points_per_node=5,
            learning_rate=0.5,
            stepwise_regression_threshold=0.01,
        )
        model.fit(datasource)
        test_preds, test_target = model.evaluate(k=3)
        # normalize outputs to be between 0 and 1
        test_preds = (test_preds - np.min(test_preds)) / (
            np.max(test_preds) - np.min(test_preds)
        )
        metrics = MetricCollection.create_from_model(model)
        results = metrics.compute(test_target, test_preds)
        # This model does not perform very well, at least
        # check that the relevant metrics are computed
        assert "MAE" in results.keys()


@unit_test
class TestMarshmallowSerialization:
    """Test Marshmallow Serialization for core Bitfount models."""

    def test_logistic_regression_serialization(
        self, datasource: BaseSource, datastructure: DataStructure
    ) -> None:
        """Test Logistic Regression serialization."""
        model = LogisticRegressionClassifier(
            datastructure=datastructure, schema=BitfountSchema()
        )
        serialized_model = bf_dump(model)
        deserialized_model = bf_load(serialized_model, MAIN_MODEL_REGISTRY)
        assert_vars_equal(vars(model), vars(deserialized_model))

    def test_regboost_regressor_serialization(
        self, datasource: BaseSource, datastructure: DataStructure
    ) -> None:
        """Test RegBoost serialization."""
        model = RegBoostRegressor(datastructure=datastructure, schema=BitfountSchema())
        serialized_model = bf_dump(model)
        deserialized_model = bf_load(serialized_model, MAIN_MODEL_REGISTRY)
        assert_vars_equal(vars(model), vars(deserialized_model))
